/* ========================================================================== *
 * Module      : Prelude.h                                                    *
 * Description : Header file for the Prelude library.                         *
 * Version     : 1.0.0 16/Jan/2004                                            *
 * ========================================================================== */

#ifndef _LIB_PRELUDE_H
#define _LIB_PRELUDE_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/* == general types ========================================================= */

typedef enum   {GT,GTE,EQL,LTE,LT} Ordering;
typedef enum   {NEQ,EQ} Eq; 
typedef void * T;
typedef enum   e_string     sel_string;
typedef struct s_string*    string;
typedef struct s_efsstring* efsstring;

/* == Primitive Types  (int, char, double) ================================== */

extern void *   int_add (void *,void *);
extern void *   int_mul (void *,void *);
extern void *   int_div (void *,void *);
extern void *   int_sub (void *,void *);
extern Ordering int_compare_int(void *,void *);
extern void *   int_succ (void *);
extern void *   int_pred (void *);
extern void *   id (void *);
extern string   bool_show    (int);
extern string   real_show    (double);
extern string   int_show     (int);
extern string   char_show    (char);
extern string   anytype_show (void *);

/* == string ================================================================ */

#define show(a) printf("%s",string_toCStr(a))

enum e_string {
   sel_string_null,
   sel_string_cons
};

struct s_string {
   sel_string tag;
   union {
      struct {
         char c;
         string next;
      } string_cons;
   } u;
};

struct s_efsstring {
 T (*string_null) (void);
 T (*string_cons) (char, T);
};

extern string     string_null     (void);
extern string     string_cons     (char, string);
extern int        free_string     (string);
extern efsstring  either_string   (T (*string_null) (void), 
                                   T (*string_cons) (char,T));
extern int        free_efsstring  (efsstring);
extern T          string_cata     (string, efsstring);
extern char *     string_toCStr   (string) ;
extern string     string_fromCStr (char *) ;
extern string     string_cat      (string,string) ;
extern int        string_eq       (string,string) ;

/* == usefull functions ===================================================== */

char * make_string(const char *fmt, ...);

#endif
/* ========================================================================== */
