/* 
 * STP - Simple Term Processor 
 * v1.0.0
 * date : 10/Feb/2004
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "lrclib.h"
#include <EVALUATOR/evaluator.h>

/* == Typedefs ============================================================== */

/* ~~ list with all output options ~~ */
typedef struct output_descriptor 
{
  struct output_descriptor *next;
  lrc_string attr;
  lrc_string view;
  lrc_string file;
  lrc_string message;
} *OutputDescriptor;


/* == Global Settings  ====================================================== */

/* ~~ default file names ~~ */
#define C_HEADER_FILE  "lib.h"
#define C_IMPL_FILE    "lib.c"
#define C_CATA_HEADER  "lib_cata.h"
#define C_CATA_IMPL    "lib_cata.c"
#define HASKELL_MODULE "lib.hs"
#define GRAPHVIZ_FILE  "lib.dot" 
#define PP_FILE        "lib.pp"
#define TECH_FILE      "lib.tt"
#define DT_DEP_FILE    "lib_dep.dot"
#define ERROR_FILE     "lib.err"

/* ~~ default options ~~ */
static int option_QUIET_MODE = 0;

/* ~~ output list (attr,view,file) ~~ */
static OutputDescriptor all_outputs = 0;

/* == Prototypes ============================================================ */

/* ~~ print ~~ */
void print_greeting_message (void);
void println_message        (char*);
void print_message          (char*);
void println_error_message  (char*);
void print_error_message    (char*);

/* ~~ options ~~ */
void explain_options        (void);
int  extract_output_file   (char*,char**);
void process_option         (char*);

/* ~~ LRC external functions ~~ */
extern lrc_term lrc_parse_file(char * fname, int nt);
extern lrc_term lrc_yacc_parse_string(char * s, int nt);
extern void lrc_set_ip_attr(lrc_term *ip, char * name, lrc_term attr_value);

/* == LRC system ============================================================ */

/*
 * read information about attribute grammar.
 */
static char* search_path[]  = { ".", "/users/matthys/d.lrc/SSL_C", 0 };
static int   have_ag_tables = 0;

#ifdef USER_ACTIONS
extern lrc_bool b_trace_equations;
#endif

#ifdef USER_ACTIONS
void init_ag_tables() {
  if (have_ag_tables) return;
  if (!lrc_read_ag_tables(search_path)) {
     printf("Could not read data of attribute grammar\n");
     exit(1);
  }
  allocate_eq_trace_data(ag_num_equations());
  have_ag_tables=1;
}
#endif
/*
 * add a triple (attr,view,file) to be processed to the list.
 */
static void enter_output_descriptor(lrc_string attr, lrc_string view, lrc_string file, lrc_string m) 
{ OutputDescriptor od;

  od = (OutputDescriptor) lrc_malloc(sizeof(*od));
  od->attr = attr; 
  od->view = view; 
  od->file = file;
  od->message = m;
  od->next = all_outputs;
  all_outputs = od; 
}
/*
 * Lrc internal function.
 */
void LrcMatchFailure() 
{ println_error_message("Matching failed!!!");
  println_error_message("this is probably fatal...\n");
}
/*
 * Lrc internal function.
 */
int lrc_printf(lrc_string s) 
{ print_message(s);
  return 1;
}
/*
 * unparse the root non terminal of the attribute grammar to the current path.
 */
static void unparse_decorated_term(DecoratedTerm dt, lrc_string view_name, lrc_string file_name) 
{ UnparseDescriptor ud;
  Reference ref;
 
  ref = create_root_reference(dt->ast);
  if (dt==0) return;
  ud = dt_sr_view2unparse_descriptor(dt,ref,view_name);
  unparse_to_file(ud,file_name);
}
/*
 * unparse the roor non-terminal of the AG to a specified path.
 */
static void unparse_dt_path(DecoratedTerm dt, lrc_string view_name, Path p, lrc_string file_name) 
{ UnparseDescriptor ud;
  Reference ref;
 
  ref = create_root_reference(dt->ast);
  ref->sub0 = p;
  
  if (dt==0) return;
  ud = dt_sr_view2unparse_descriptor(dt,ref,view_name);
  unparse_to_file(ud,file_name);
}
/*
 * unparse a non terminal to a file.
 */
static void emit_unattributed_term(lrc_term t, lrc_string view_name, FILE *f) 
{ UnparseDescriptor ud;

  ud.term = t;
  ud.uat = 0;
  ud.view_name = view_name;
  unparse_to_file_stream(ud,f);
}
/*
 * collect an root attribute from a the root non-terminal.
 */
static lrc_term get_root_op_attr(char * attr_name, lrc_term *op)
{ int attr_index;

  attr_index = lrc_attr_get_op_index(lrc_root_nonterminal(), 1 /* vis_num */, attr_name);
  if (attr_index == -1) 
    {  print_error_message("INTERNAL ERROR: get_root_op_attr: could not locate attribute ");
       println_error_message(attr_name);
       return 0;
    }
  return op[attr_index];
}
/*
 * Unparse a named root attribute to a file. 
 */
static void emit_root_attribute(DecoratedTerm dt, lrc_string attr_name, lrc_string view, FILE *f) 
{ lrc_term attr;
  lrc_term * op;

  op  = dt_get_decoration(dt);
  if (op == 0) return;

  attr = get_root_op_attr(attr_name,op);
  if (attr != 0) 
  {  emit_unattributed_term(attr,view,f);  }
}
/*
 * produce the file containing the decorated term with the specified view.
 */
static int emit_according_to_output_descriptor(DecoratedTerm dt, OutputDescriptor od) 
{ FILE *f;

  /* check if there is a file name in descriptor */
  if (od->file==0) 
  {  /* if not then the output is the standard output */
     f = stdout; 
  }
  else 
  {  /* if yes, open for wrting the specified file */
     f=fopen(od->file,"w");
     if (!f) 
      {  /* if there was an error, print a error message and return */
         print_error_message("Cannot open ");
         println_error_message(od->file);
         return 0;
      }
  }

  /* check if there is a attribute name in descriptor */
  if (od->attr!=0) 
  { /* if so, produce the file, with the attribute and view specified */
    emit_root_attribute(dt,od->attr,od->view,f);
  }
  else 
  { /* if not, unparse the root non terminal with the specified view */
    unparse_decorated_term(dt,od->view,od->file);
  }
}
/*
 * compute the abstract tree based on the concrete one.
 */
lrc_term compute_abstract_tree(lrc_term concrete_tree) 
{ lrc_term * op;

  op = lrc_visit_term_1(concrete_tree);
  if (op == 0) return 0;
  return op[0];
}
/*
 * if an parse error occurs, explain where.
 */
void explain_parse_error(void) 
{
  print_message("Parser failed: ");
  switch (lrc_parse_state()) 
  {
     case PS_OK               : break;
     case PS_FILE_NOT_READABLE: println_message("file is not readable"); 
                                break;
     case PS_NO_ENTRY         : println_message("nonterminal has no parse entry"); 
                                break;
     case PS_SYNTAX_ERROR     : println_message("syntax error."); 
                                break; 
  }
}

/*
 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
 ||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
 */

/* 
 * parse a string using the concrete grammar defined in SSL.
 */
lrc_term string_to_concrete(char * s, int nt) 
{ lrc_term parse_tree = lrc_yacc_parse_string(s,nt);
  if (parse_tree == 0) explain_parse_error();
  return parse_tree;
}
/*
 * parse a file using the concrete grammar defined in SSL.
 */
lrc_term file_to_concrete(char * fname,int nt) 
{ lrc_term parse_tree = lrc_parse_file(fname,nt);
  if (parse_tree == 0) explain_parse_error();
  return parse_tree;
}
/*
 * convert concrete term to an abstract one.
 */
lrc_term concrete_to_abstract(lrc_term concrete_tree) 
{ lrc_term * op;
 
  if (concrete_tree == 0) return 0;
  lrc_memo_start_AE();
  op = lrc_visit_term_1(concrete_tree);
  lrc_memo_end_AE();
  if (op == 0) return 0;
  return op[0];
}
/* 
 * decorate the abstract tree, evaluate the grammar.
 */
DecoratedTerm decorate_ast(lrc_term abs_tree) 
{ DecoratedTerm res;
  
  if (abs_tree == 0) return 0;
  res = create_decorated_term();
  dt_enter_ast(res,abs_tree);
  dt_decorate_term(res);
  return res;
}
/*
 * process the decorated term, producing the files according
 * to the information in the Output Descriptor. 
 */
void process_decorated_term(DecoratedTerm dt) 
{ lrc_term ast;
  lrc_term * op;
  
  if (dt == 0) return;
  ast = dt_edse_initial_ast(dt);
  op  = dt_get_decoration(dt);

  if (op == 0 || ast == 0) return;
 
  if (all_outputs != 0) 
   {  OutputDescriptor od;
      od = all_outputs;
      while (od != 0) 
      {  /* print message if any..*/
         print_message(od->message);
         emit_according_to_output_descriptor(dt,od);
         od=od->next;
      }
   }
}
/*
 *
 */
void process_abstract_tree(lrc_term ast) 
{ if (ast==0) return;
  process_decorated_term(decorate_ast(ast));
}
/*
 * process a input file, parsing it, building the abstract syntax tree,
 * decorating it, and build the files according to the Output Descriptor.
 */
static int lrc_process_file(char * fname, int nt) 
{ lrc_term concrete;
  lrc_term abstract;

  concrete = file_to_concrete(fname,nt);
  if (concrete == 0) return 0;
  abstract = concrete_to_abstract(concrete);
  if (abstract == 0) return 0;
  process_abstract_tree(abstract);
  return 1;
}
/*
 * process a string parsing, building the ast and decorating it, producing
 * the files specified in the Output Descriptor.
 */
void lrc_process_string(char * str,int nt) 
{  process_abstract_tree(concrete_to_abstract(string_to_concrete(str,nt))); }
/*
 * initialize LRC system.
 */
static void Init(void) 
{
  lrc_lib_init();
  lrc_memo_init();
  init_grammar_tables();
  init_unparser();
  /*** lrc_set_string_parser(lrc_yacc_parse_string); **/
}

/* == STP system ======================================================================== */

/*
 * display available options.
 */
void explain_options(void) 
{ println_message ("Usage: stp [OPTIONS...] INPUT_FILE");
  println_message ("");
  println_message ("  -C[OUTPUT_FILE]  produce C code");
  println_message ("  -H[OUTPUT_FILE]  produce Haskell code");
  println_message ("  -G[OUTPUT_FILE]  produce GraphViz file");
  println_message ("  -P[OUTPUT_FILE]  produce a pretty printing representation of the input");
  println_message ("  -T[OUTPUT_FILE]  produce technical information about grammar (debug purposes)");
  println_message ("  -D[OUTPUT_FILE]  produce a GraphViz file with the datatypes dependencies");
  println_message ("  -w               list errors/warnings of specification");
  println_message ("  -y               trace parsing");
  println_message ("  -q               quiet mode");
  println_message ("  -h               print this message");
  println_message ("");
  println_message ("with no OPTIONS, produces C code.");
  println_message ("with no OUTPUT_FILE produces default file names.");
}
/*
 * extract output filename from the command line option.
 */
int extract_output_file(char *str, char**out)
{ char * outfile;
  int  exist_out_file = 0;

  outfile = (char *) lrc_malloc(strlen(str) + 1);
  /* skip leading - and option tag, ex C,D,P... */
  outfile = strcpy(outfile,str);
  outfile += 2;
  /* verify if there is a output file */
  exist_out_file = (strlen(str) > 2);
  *out = outfile;  
  
  return (exist_out_file);
}
/*
 * process an option, determining what to do.
 */
void process_option(char *str) 
{ char *output_file ;

  switch(str[1]) 
  {
        /* -C produce C header & implementation */
        case 'C' : if (extract_output_file(str,&output_file))
                   { /* append the .h and .c extension to the file */
                     char *header_file = lrc_vstring_concat(output_file,".h",0);
                     char *impl_file   = lrc_vstring_concat(output_file,".c",0);  
                     char *cata_header = lrc_vstring_concat(output_file,"_cata.h",0);
                     char *cata_impl   = lrc_vstring_concat(output_file,"_cata.c",0);
                     /* append to the output descriptor */
                     enter_output_descriptor("c_libheader","BASEVIEW",header_file,"");  
                     enter_output_descriptor("c_libimpl","BASEVIEW",impl_file,"");
                     enter_output_descriptor("c_cataheader","BASEVIEW",cata_header,"");
                     enter_output_descriptor("c_cataimpl","BASEVIEW",cata_impl,"");
                   }
                   else
                   { /* append to the output descriptor default file names */
                     enter_output_descriptor("c_libheader","BASEVIEW",C_HEADER_FILE,"");
                     enter_output_descriptor("c_libimpl","BASEVIEW",C_IMPL_FILE,"");
                     enter_output_descriptor("c_cataheader","BASEVIEW",C_CATA_HEADER,"");
                     enter_output_descriptor("c_cataimpl","BASEVIEW",C_CATA_IMPL,"");
                   }
                   break;
        /* -H produce Haskell module */
        case 'H' : if (extract_output_file(str,&output_file))
                   { /* append the .hs extension to the output file */
                     char *haskell_module = lrc_vstring_concat(output_file,".hs",0);
                     /* append to the output descriptor */
                     enter_output_descriptor("h_module","BASEVIEW",haskell_module,"");
                   }
                   else
                   { /* use default name for haskell modules */
                     enter_output_descriptor("h_module","BASEVIEW",HASKELL_MODULE,"");
                   }
                   break;
        /* -G produce Graphviz file */
        case 'G' : if (extract_output_file(str,&output_file))
                   { /* append the .dot extension to the output file */
                     char *graphviz_file = lrc_vstring_concat(output_file,".dot",0);
                     /* append to the output descriptor */
                     enter_output_descriptor(0,"GraphViz",graphviz_file,"");
                   }
                   else
                   { /* use default */
                     enter_output_descriptor(0,"GraphViz",GRAPHVIZ_FILE,"");
                   }
                   break;
        /* -P produce pretty printing */
        case 'P' : if (extract_output_file(str,&output_file))
                   { /* append the .pp extension to the output file */
                     char *prettyp_file = lrc_vstring_concat(output_file,".pp",0);
                     /* append to the output descriptor */
                     enter_output_descriptor("dag_pp","BASEVIEW",prettyp_file,"");
                   }
                   else
                   { /* default name */
                     enter_output_descriptor("dag_pp","BASEVIEW",PP_FILE,"");
                   }
                   break;
        /* -T techinal information */
        case 'T' : if (extract_output_file(str,&output_file))
                   { /* append the .tt extension to the output file */
                     char *tech_file = lrc_vstring_concat(output_file,".tt",0);
                     /* append to the output descriptor */
                     enter_output_descriptor("dag","BASEVIEW",tech_file,"");
                   }
                   else
                   { enter_output_descriptor("dag","BASEVIEW",TECH_FILE,"");
                   }
                   break;
        /* -D produce dependecies */
        case 'D' : if (extract_output_file(str,&output_file))
                   { /* append the .dot extension to the output file */
                     char *dp_dep_file = lrc_vstring_concat(output_file,"_dep.dot",0);
                     /* append to the output descriptor */
                     enter_output_descriptor("nt_inter_dep","GraphViz",dp_dep_file,"");
                   }
                    else
                      enter_output_descriptor("nt_inter_dep","GraphViz",DT_DEP_FILE,"");
                   break;
        /* -w produce warnings */
        case 'w' : enter_output_descriptor(0,"SEM_ERROR",ERROR_FILE,"");
                   enter_output_descriptor("error_list","BASEVIEW",0,"validating specification...\n");
                   break;           
        /* -y trace parsing */
        case 'y' : set_parse_debug(1);
                   break;
        /* -h help on options */
        case 'h' : explain_options();
                   exit(1);
                   break;
        /* -q quiet mode */
        case 'q' : option_QUIET_MODE = 1;
                   break;
        /* not a valid option */ 
        default  : print_error_message("Unknown option, ");
                   print_error_message(str);
                   println_error_message(", use -h to display options.");
		   exit(1);
  } 
}
/* 
 * print greeting message. 
 */
void print_greeting_message(void)
{ println_message(" ___  ____  ___    _______________________________________");
  println_message("||     ||  |   ||  Stp : Simple Term Processor            ");
  println_message("||__   ||  |___||  (powered by LRC)                       "); 
  println_message("   ||  ||  ||      URL:                                   ");
  println_message(" __||  ||  ||      Bugs, sugestions to :                  ");
  println_message(" Version 1.0.0     _______________________________________");
}
/*
 * print a message to the standard output ataching a newline carater.
 */
void println_message(char *s)
{ print_message(s);
  print_message("\n"); }
/*
 * print a message to the standard output.
 */
void print_message(char *s)
{ if (!option_QUIET_MODE) fprintf(stdout,"%s",s); }
/*
 * print a error message to the standard error output ataching a newline.
 */
void println_error_message(char *s)
{ print_error_message(s);
  print_error_message("\n");
}
/* 
 * print a error message.
 */
void print_error_message(char *s)
{ if (!option_QUIET_MODE) fprintf(stderr,"%s",s); }

/*============================================================================*  
 * MAIN                                                                       *   
 *============================================================================*/  
int main(int argc, char **argv)
{ int at_least_one_option = 0;
  int arg_num;

  if (argc == 1) 
  {  /* if no parameters introduced, display help message and exit */
     println_message("use option -h to view available options.");
     return 1;  
  }

  /* initialize LRC system */
  Init();

  /* process options */
  for (arg_num=1;arg_num<argc;arg_num++) 
    {  if (argv[arg_num][0]=='-') 
       {  /* an option */
          process_option(argv[arg_num]);
          at_least_one_option += 1;
       }
    }

  /* check to see if there is an input file */
  if (at_least_one_option == (argc - 1) )
  { println_message("no input file. use -h for some help.");
    return 1;
  }

  print_greeting_message();
             
  /* if no options produce C code */
  if (!at_least_one_option) 
    process_option("-C");

  /* after reading options, if any, process the input file */
  print_message("stp : input file [");
  print_message(argv[at_least_one_option + 1]);
  println_message("]");
  lrc_process_file(argv[at_least_one_option + 1],lrc_root_nonterminal());

  return 0;
}
/* ========================================================================== */

