#include <stdio.h>
#include "lrcassert.h"
#include "lrcmemory.h"
#include "prims.h"
#include "term.h"
#include "initial.h"
#include "grammar_tables.h"
#include "ebinding.h"
#include "reference.h"
#include "decoration.h"
#include "surgery.h"
#include "binding_event.h"

#ifndef TRACE_HANDLING
#define TRACE_HANDLING 0
#endif


static void print_match_parameters(lrc_term _lrc_ptrs) {
  while (lrc_constructor_num_sons(LRC_TAG(_lrc_ptrs))==2) {
    lrc_term ptr;
    ptr=LRC_CHILD(_lrc_ptrs,1);
    printf("    PTR %p:",ptr); lrc_term_print_long(ptr); printf("\n");
    _lrc_ptrs=LRC_CHILD(_lrc_ptrs,2);
  }
}

static void print_binding_instance(lrc_term binding_instance) {
  int binding_number=LRC_VALUE_int(LRC_CHILD(binding_instance,1));
  lrc_term match_parameters=LRC_CHILD(binding_instance,2);
  lrc_term trafo_data=LRC_CHILD(binding_instance,3);

  printf("BINDING INSTANCE of %d\n", binding_number);
  printf("  MATCH_PARAMETERS\n"); print_match_parameters(match_parameters);
  printf("  TRAFO_DATA"); lrc_term_print_long(trafo_data); printf("\n");  
}

static void print_binding_instances(lrc_term binding_instances) {
  while (lrc_constructor_num_sons(LRC_TAG(binding_instances))==2) {
    print_binding_instance(LRC_CHILD(binding_instances,1));
    binding_instances=LRC_CHILD(binding_instances,2);
  }
}


int dt_handle_abstract_event(DecoratedTerm dt, lrc_term event, lrc_term frame) {
  lrc_term bindings;
  lrc_term new_term;
  int result;
  Replacement replacement;

  /*
  ** Determine whether a binding exists for the name of the visual object 
  ** that is contained in the event. (always the first subterm)
  */
  new_term=dt->ast;
  result=NOT_HANDLED | NOT_CHANGED_TREE;
  if (TRACE_HANDLING) {
    printf("TRYING BINDING INSTANCES\n"); print_binding_instances(dt->derived_bindings);
    printf("\n");
  }
  bindings=lrc_all_matches(event,frame,dt->derived_bindings);

  if ( is_empty_bindings(bindings) ) { 
     if (TRACE_HANDLING) { printf(">> NO MATCHING BINDING\n"); }
     return 0;
  }
  if ( !is_empty_bindings(bindings) ) { /* A binding exists for the visual object. */
    lrc_term first_binding;
    int binding_number;
    lrc_term trafo_datas;
    int num_trafos, trafo_number;
    Reference * operand_references;
    lrc_term * replacements;

    if (TRACE_HANDLING) {
      printf(">> HABEMUS MATCHES !!\n");
    }
    first_binding=LRC_CHILD(bindings,1); /* :: _lrc_binding_instance */

    binding_number=instance_binding_number(first_binding);
    trafo_datas=instance_trafo_data(first_binding);
    
    num_trafos=binding_num_trafos(binding_number);
    replacements=(lrc_term *)lrc_malloc(num_trafos * sizeof(lrc_term) );
    operand_references=(Reference *)lrc_malloc(num_trafos * sizeof(*operand_references));
    replacement=create_replacement(dt->ast, num_trafos);

    trafo_number=0;
    while (!is_empty_trafo_data(trafo_datas)) {
       lrc_term td;
       lrc_term parameters;
       lrc_term operand;

       td=LRC_CHILD(trafo_datas,1);
       parameters=bind_trafo_parameters(td);

       replacement->replacements[trafo_number]
       = invoke_transformation(binding_number,trafo_number,event,frame,parameters);

       replacements[trafo_number]
        = invoke_transformation(binding_number,trafo_number,event,frame,parameters);
       operand=bind_trafo_subtree_reference(td);
       if (TRACE_HANDLING) {
         printf("OPERAND==");
         if (LRC_TAG(operand)==lrc_prod_name_to_tag("_Ptr")) {
           printf("->"); lrc_term_print_long(LRC_VALUE_lrc_ptr(operand));
         }
         else {
           lrc_term_print_long(operand);
         }
         printf("\n");
       }

       operand_references[trafo_number]
        = term_to_reference(new_term,operand);
       
       replacement->operands[trafo_number]=term_to_reference(new_term,operand);

       if (TRACE_HANDLING) {
         printf("FILLING BLOCK ELEMENT %d\n", trafo_number);
         printf("OPERAND REFERENCE = ");
           path_print(operand_references[trafo_number]->sub0); printf("\n");
         printf("REPLACEMENT = ");
           lrc_term_print_long(replacements[trafo_number]); printf("\n");
       }

       trafo_datas=LRC_CHILD(trafo_datas,2);
       trafo_number++;
    }

    new_term=in_situ_multi_replace(new_term, replacement);

    if (TRACE_HANDLING) {
      printf("REPLACEMENT=..."); /** lrc_term_print_indent(new_term,1); **/
      printf("\n");
    }
    result=result|FULLY_HANDLED;
    /* printf("HANDLED A BINDING: result==%d\n",result); */
  }
  if (lrc_term_equal(new_term,dt->ast)) {
    result=(result&~CHANGED_TREE);
    /* printf("TREE HAS NOT CHANGED: result==%d\n",result); */
  } 
  else {
    result=(result|CHANGED_TREE);
    /* printf("TREE HAS CHANGED: result==%d\n",result); */
  }
  dt_enter_ast(dt,new_term);
  return result;
}

