#include <stdio.h>
#include "lrcassert.h"
#include "lrcmemory.h"
#include "prims.h"
#include "term.h"
#include "initial.h"
#include "reference.h"
#include "grammar_tables.h"
#include "visit_tables.h"
#include "decoration.h"

static lrc_term empty_derived_bindings(void) {
  if ( lrc_prod_name_to_tag("Derived_BindingsNil")==-1 ) {
    return 0;
  }
  return lrc_build_term("Derived_BindingsNil",0);
}

DecoratedTerm create_decorated_term() {
  DecoratedTerm res;
  res=(DecoratedTerm)lrc_malloc(sizeof(*res));
  res->ast_state=NO_AST;
  return res;
}


static ast_state_tp set_ast(DecoratedTerm dt, lrc_term new_term) {
  if (dt->ast_state==NO_AST) {
    dt->ast_state=NEW_BORN_AST;
  }
  else if (!lrc_term_equal(dt->ast,new_term)) { dt->ast_state=DIFF_FROM_PREV_AST; }
  else  { dt->ast_state=SAME_AS_PREV_AST; }

  if (dt->ast_state!=SAME_AS_PREV_AST) dt->ast=new_term;

   return dt->ast_state;
}

void dt_enter_ast(DecoratedTerm dt, lrc_term abstract_tree) {
  ast_state_tp state;
  /* The initial AST is an initial (???) term of the abstract tree. */
  /*****
    Assert(lrc_is_initial(abstract_tree),"dt_enter_ast: parameter abstract_tree is not initial");
  ******/

  /***
    printf("ENTERING AST to Decorated Term\n");
    printf("AST="); lrc_term_print_long(abstract_tree); printf("\n");
  ***/
  state=set_ast(dt, abstract_tree);
  if (dt->ast_state!=SAME_AS_PREV_AST) {
    /* There are no derived bindings yet. */
    dt->derived_bindings=empty_derived_bindings();
  }
}

/* This function simply returns the current initial AST */
lrc_term dt_edse_initial_ast(DecoratedTerm dt) {
  return dt->ast;
}

void  dt_decorate_term(DecoratedTerm dt) {
  lrc_term *decoration;

  decoration=lrc_visit_edse_root(dt->ast, empty_derived_bindings());
  dt->decoration=decoration;
}

lrc_term * dt_get_decoration(DecoratedTerm dt) {
  return dt->decoration;
}


void dt_install_root_attr(DecoratedTerm dt,lrc_term ast, lrc_term root_attr, char * name) {
  if (   lrc_prod_lhs_nont(LRC_TAG(root_attr))
       ==lrc_id_to_nont("_lrc_bindings") ) {
    /* printf("dt_install_root_attr: Setting derived bindings!!!!!\n"); */
    dt->derived_bindings=root_attr;
  }
}


