#include <string.h>
#include <stdio.h>
#include "shalloc.h"
#include "initial.h"
#include "prims.h"
#include "strs.h"
#include "textbuffer.h"
#include "position.h"
#include "unparse_tables.h"
#include "view_tables.h"
#include "reference.h"
#include "fragments.h"
#include "views.h"
#include "ints.h"
#include "reals.h"
#include "maps.h"
#include "style.h"
#include "foreign.h"
#include "grammar_tables.h"
#include "decoration.h"
#include "unparser.h"
#include "file_unparser.h"

#ifndef TRACE_UNPARSER
#define TRACE_UNPARSER 0
#endif

#define TRACE_START if(TRACE_UNPARSER) {
#define TRACE_END   }



#define EMIT_CHAR(ch) fprintf(OUT_FILE,"%c",ch)
#define EMIT_NL() fprintf(OUT_FILE,"\n")

typedef struct style_stack_element {
  struct style_stack_element * below;
  lrc_string style_name;
  Position start_position; /* Where does a range of chars in this style start */
} StyleStackElement, * StyleStack;


static FILE * OUT_FILE;

static void emit(char * s, int offset, int num_chars) {
  int six;
  for (six=0;six<num_chars;six++) {
     EMIT_CHAR(s[offset+six]);
  }
}

static void emit_n_char(int num_chars, char c) {
  int six;
  for (six=0;six<num_chars;six++) {
     EMIT_CHAR(c);
  }
}

static StyleStack push_style(lrc_string style_name, StyleStack ss) {
  StyleStack res;

  res=(StyleStack)lrc_malloc(sizeof(*res));
  res->below=ss;
  res->style_name=style_name;
  return res;
}

static StyleStack pop_style(StyleStack ss) {

  if (ss==0) return 0;
  return ss->below;
}

#define ELEMENT bh->elements[index]
void fragments_to_file_stream(ViewInstance vi, FILE *f) {
  Fragments fr;
  UnparseBlockHeader bh;
  unparse_commands_tp * commands;
  Position write_position;
  int indentation;
  StyleStack style_stack;
  p_lrc_string_list configured_styles;


  OUT_FILE=f;
  fr=vi->fragments;
  if (fr==0) return;
  commands=vi->view_table.view_unparse_commands;
  write_position.row=1;
  write_position.col=0;
  indentation=0;
  style_stack=0;
  style_stack=push_style("BASESTYLE",style_stack);
  style_stack->start_position=write_position;
  configured_styles=0;

  for (bh=fr->first;bh!=0; bh=bh->next) {
     int index;
     for (index=0; index<bh->free_index; index++) {
       switch (ELEMENT.kind) {
         case NO_FRAGMENT : 
; break;
         case TEMPLATE_STR_REF_FRAGMENT : {
                                            int offset, num_chars;
                                            char * str;
                                            offset=TEMPLATE_STR_REF(ELEMENT).offset;
                                            num_chars=TEMPLATE_STR_REF(ELEMENT).num_chars;
                                            str=(*commands)[TEMPLATE_STR_REF(ELEMENT).template_index].command_data;

                                            emit(str,offset,num_chars);
                                            write_position.col+=num_chars;
                                          }
; break;
         case TREE_STR_REF_FRAGMENT : { lrc_string s;
                                        s=LRC_VALUE_lrc_string(TREE_STRING_REF(ELEMENT).term);
                                        
                                        emit(s,TREE_STRING_REF(ELEMENT).offset,TREE_STRING_REF(ELEMENT).num_chars );
                                        write_position.col+=TREE_STRING_REF(ELEMENT).num_chars;
                                      }
; break;
         case TREE_CHAR_FRAGMENT : 
                                   EMIT_CHAR( LRC_VALUE_char(TERM(ELEMENT)) );
                                   write_position.col++;
; break;
         case CONSTANT_STR_FRAGMENT : 
                                      emit(STRING(ELEMENT), 0, strlen(STRING(ELEMENT)));
                                      write_position.col+=strlen(STRING(ELEMENT));
; break;
         case HARD_NL_FRAGMENT : 
                                 
                                 EMIT_NL();
                                 write_position.row++;
                                 write_position.col=0;
                                 
                                 emit_n_char(indentation, ' ');
                                 write_position.col+=indentation;

; break;
         case OPTIONAL_NL_FRAGMENT : 
; break;
         case CONDITIONAL_NL_FRAGMENT : 
; break;
         case INDENT_UP_FRAGMENT : indentation=indentation+2;
; break;
         case INDENT_DOWN_FRAGMENT : indentation=indentation-2;
                                     if (indentation<0) indentation=0;
; break;
         case OPEN_GROUP_FRAGMENT : 
; break;
         case CLOSE_GROUP_FRAGMENT : 
; break;
         case PUSH_STYLE_FRAGMENT : 
                                    style_stack=push_style(STRING(ELEMENT),style_stack);
                                    if (style_stack!=0) style_stack->start_position=write_position;
; break;
         case POP_STYLE_FRAGMENT : 
                                   style_stack=pop_style(style_stack);
                                   if (style_stack!=0) style_stack->start_position=write_position;
; break;
         case CHAR_FRAGMENT : 
                              EMIT_CHAR( CHAR(ELEMENT) );
                              write_position.col++;
; break;
         case ORIGIN_FRAGMENT : 

; break;
         case UP_FRAGMENT: break;
         case CHILD_FRAGMENT: break;
         case ENTER_ATTR_FRAGMENT: break;
       }

     }
  }
  
}

void unparse_to_file_stream(UnparseDescriptor ud, FILE *f) {
  ViewInstance vi;

  vi=unparse_descriptor_to_vi(ud);
  vi_create_fragments(vi);
  TRACE_START
    vi_print(vi);
  TRACE_END
  fragments_to_file_stream(vi,f);
}

void unparse_to_file(UnparseDescriptor ud, lrc_string file_name) {
  FILE *f;
  if (file_name==0) {
    f=stdout;
  }
  else {
    f=fopen(file_name,"w");
    if (f==0) {
      printf("Could not open %s\n",file_name);
      return;
    }
  }

  unparse_to_file_stream(ud,f);
  fclose(f);
}


