#include <stdio.h>
#include <string.h>
#include "shalloc.h"
#include "initial.h"
#include "prims.h"
#include "strs.h"
#include "textbuffer.h"
#include "position.h"
#include "unparse_tables.h"
#include "view_tables.h"
#include "reference.h"
#include "position.h"
#include "fragments.h"

#ifndef TRACE_UNPARSER
#define TRACE_UNPARSER 0
#endif

#define TRACE_START if (TRACE_UNPARSER) {
#define TRACE_END }

#ifndef TRACE_POINTING
#define TRACE_POINTING 0
#endif

#define TRACE_POINTING_START if (TRACE_POINTING) {
#define TRACE_POINTING_END }



void print_unparse_element_kind(UnparseElementKind k) {
  switch(k) {
    case NO_FRAGMENT               : printf("NONE"); break;
    case CHILD_FRAGMENT            : printf("CHILD"); break;
    case UP_FRAGMENT               : printf("UP"); break;
    case ENTER_ATTR_FRAGMENT       : printf("ENTER_ATTR"); break;
    case TEMPLATE_STR_REF_FRAGMENT : printf("STR_REF"); break;
    case TREE_STR_REF_FRAGMENT     : printf("TREE_STR_REF"); break;
    case TREE_CHAR_FRAGMENT        : printf("TREE_CHAR"); break;
    case CONSTANT_STR_FRAGMENT     : printf("CONSTANT_STR"); break;
    case HARD_NL_FRAGMENT          : printf("NL"); break;
    case OPTIONAL_NL_FRAGMENT      : printf("OPTIONAL_NL"); break;
    case CONDITIONAL_NL_FRAGMENT   : printf("CONDITIONAL_NL"); break;
    case INDENT_UP_FRAGMENT        : printf("INDENT_UP"); break;
    case INDENT_DOWN_FRAGMENT      : printf("INDENT_DOWN"); break;
    case OPEN_GROUP_FRAGMENT       : printf("OPEN"); break;
    case CLOSE_GROUP_FRAGMENT      : printf("CLOSE"); break;
    case PUSH_STYLE_FRAGMENT       : printf("PUSH"); break;
    case POP_STYLE_FRAGMENT        : printf("POP"); break;
    case SET_INDENT_FRAGMENT       : printf("SET_INDENT"); break;
    case CHAR_FRAGMENT             : printf("CHAR"); break;
    case ORIGIN_FRAGMENT           : printf("ORIGIN"); break;
  }
}

void print_unparse_element(UnparseElement e, unparse_commands_tp * commands) {
  print_unparse_element_kind(e.kind);
  if (e.kind==TEMPLATE_STR_REF_FRAGMENT) {
    int offset, num_chars, ix;
    char * str;
    offset=TEMPLATE_STR_REF(e).offset;
    num_chars=TEMPLATE_STR_REF(e).num_chars;
    printf(":[%d]+%d:%d",
            TEMPLATE_STR_REF(e).template_index,
            offset,
            num_chars);
    str=(*commands)[TEMPLATE_STR_REF(e).template_index].command_data;
    printf("[%s] |-> <",str);
    for(ix=0;ix<num_chars;ix++) {
      printf("%c",str[offset+ix]);
    }
    printf(">");
  }
  if (e.kind==CHILD_FRAGMENT) {
    printf(":<%d>", INT(e));
  }
  if (e.kind==CONSTANT_STR_FRAGMENT || e.kind==PUSH_STYLE_FRAGMENT) {
    printf(":<%s>", STRING(e));
  }
  if (e.kind==TREE_STR_REF_FRAGMENT) {
    lrc_term_print_long(TREE_STRING_REF(e).term);
    printf("+%d:%d",TREE_STRING_REF(e).offset,TREE_STRING_REF(e).num_chars);
  }
  if (e.kind==TREE_CHAR_FRAGMENT) {
    lrc_term_print_long(TERM(e));
    printf(":<%c>",LRC_VALUE_char(TERM(e)));
  }
  if (e.kind==CHAR_FRAGMENT) {
    printf(":<%c>",CHAR(e));
  }
  if (e.kind==ORIGIN_FRAGMENT) {
    printf(":&%p",TERM(e));
  }
}
void fragments_print_block_info(Fragments f) {
  UnparseBlockHeader bh;
  int block_num=0;

  if (f==0) return;
  for (bh=f->first;bh!=0; bh=bh->next) {
     printf("BLOCK %d: num_elements=%d; free_index=%d\n",
            block_num, bh->num_elements, bh->free_index);
     block_num++;
  }
}
UnparseStack push_unparse(int indentation, UnparseStack us) {
  UnparseStack res;
  res=(UnparseStack)lrc_malloc(sizeof(*res));
  res->below=us;
  res->indentation=indentation;
  res->indentation_string=0;
  return res;
}

UnparseStack pop_unparse(UnparseStack us) {
  if (us==0) return 0;
  return us->below;
}

static UnparseBlockHeader create_block_header(int num_elements) {
  UnparseBlock elements;
  UnparseBlockHeader res;

  /* printf("create_block(_header) with %d bytes (%d elements)\n",
                 num_elements * sizeof(*elements), num_elements ); */
  elements=(UnparseBlock)lrc_malloc(num_elements * sizeof(*elements));
  res=(UnparseBlockHeader)lrc_malloc(sizeof(*res));
  res->next=0;
  res->prev=0;
  res->num_elements=num_elements;
  res->free_index=0;
  res->elements=elements;
  return res;   
}

Fragments create_fragments(void) {
  Fragments res;
  res=(Fragments)lrc_malloc(sizeof(*res));
  res->first=0;
  res->last=0;
  return res;
}

void fragments_delete_elements(Fragments fr) {
  fr->first=0;
  fr->last=0;
}

static void fragments_add_block_header(Fragments f, UnparseBlockHeader bh) {
  if (f->first==0) {
    f->first=bh;
    f->last=bh;
    return;
  }
  bh->prev=f->last;
  f->last->next=bh;
  f->last=bh;
}

static int fragments_available_space(Fragments f) {
  if (f->first==0) return 0;
  return f->last->num_elements - f->last->free_index;
}

static void fragments_extend(Fragments f, int min_needed) {
  if (min_needed<MIN_UNPARSE_BLOCK_SIZE) min_needed=MIN_UNPARSE_BLOCK_SIZE;
  fragments_add_block_header(f,create_block_header(min_needed));
}

static void assure_space(Fragments f, int needed) {
  if (needed>fragments_available_space(f)) fragments_extend(f,needed);
}

void fragments_add_element(Fragments f, UnparseElement e) {
  assure_space(f,1);
  f->last->elements[f->last->free_index]=e;
  f->last->free_index=f->last->free_index + 1;
}

void fragments_all_elements(Fragments f,  void action()) {
  UnparseBlockHeader bh;

  if (f==0) return;

  for (bh=f->first;bh!=0; bh=bh->next) {
     int index;
     for (index=0; index<bh->free_index; index++) {
       action(bh->elements[index]);
     }
  }
}



void fr_add_template_substring(int template_index, int char_offset, int length, Fragments f) {
  UnparseElement e;
  
  e.kind=TEMPLATE_STR_REF_FRAGMENT;
  e.element_data.tmpl_sr.template_index=template_index;
  e.element_data.tmpl_sr.offset=char_offset;
  e.element_data.tmpl_sr.num_chars=length;
  fragments_add_element(f,e);
}

void fr_add_tree_substring(lrc_term t, int char_offset, int length, Fragments f){
  assure_space(f,1);
  FREE_ELEMENT(f).kind=TREE_STR_REF_FRAGMENT;
  TREE_STRING_REF(FREE_ELEMENT(f)).term=t;
  TREE_STRING_REF(FREE_ELEMENT(f)).offset=char_offset;
  TREE_STRING_REF(FREE_ELEMENT(f)).num_chars=length;
  BUMP(f);
}

void fr_add_tree_char(lrc_term t, Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=TREE_CHAR_FRAGMENT;
  TERM(FREE_ELEMENT(f))=t;
  BUMP(f);
}

void fr_add_constant_string(char * str, Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=CONSTANT_STR_FRAGMENT;
  STRING(FREE_ELEMENT(f))=str;
  BUMP(f);
}

void fr_add_indent_up(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=INDENT_UP_FRAGMENT;
  BUMP(f);
}

void fr_add_indent_down(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=INDENT_DOWN_FRAGMENT;
  BUMP(f);
}


void fr_add_set_indent(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=SET_INDENT_FRAGMENT;
  BUMP(f);
}

void fr_add_line_break(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=HARD_NL_FRAGMENT;
  BUMP(f);
}

void fr_add_optional_line_break(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=OPTIONAL_NL_FRAGMENT;
  BUMP(f);
}

void fr_add_consistent_line_break(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=CONDITIONAL_NL_FRAGMENT;
  BUMP(f);
}

void fr_add_open_group(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=OPEN_GROUP_FRAGMENT;
  BUMP(f);
}

void fr_add_close_group(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=CLOSE_GROUP_FRAGMENT;
  BUMP(f);
}

void fr_add_constant_char(char c, Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=CHAR_FRAGMENT;
  CHAR(FREE_ELEMENT(f))=c;
  BUMP(f);
}

void fr_add_push_style(char * style_name, int style_name_length, Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=PUSH_STYLE_FRAGMENT;
  STRING(FREE_ELEMENT(f))=style_name;
  BUMP(f);
}

void fr_add_pop_style(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=POP_STYLE_FRAGMENT;
  BUMP(f);
}

void fr_add_origin(lrc_term t, Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=ORIGIN_FRAGMENT;
  TERM(FREE_ELEMENT(f))=t;
  BUMP(f);
}

void fr_add_child(int n, Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=CHILD_FRAGMENT;
  INT(FREE_ELEMENT(f))=n;
  BUMP(f);
}

void fr_add_up(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=UP_FRAGMENT;
  BUMP(f);
}

void fr_add_enter_attr(Fragments f) {
  assure_space(f,1);
  FREE_ELEMENT(f).kind=ENTER_ATTR_FRAGMENT;
  BUMP(f);
}


static int unparse_ref_equal(UnparseElementReference r0, UnparseElementReference r1) {
  if (r0.f!=r1.f) return 0;
  if (r0.f==0) return 1;
  if (r0.block!=r1.block) return 0;
  return r0.index==r1.index;
}


static void print_unparse_element_reference(UnparseElementReference uer) {
  if (uer.f==0) { printf("<null>"); return; }
  if (uer.block==0) { printf("<null_block>"); }
  printf("%d",uer.index);
}

#define print_bool(b) if (b) printf("Y"); else printf("N")
#define AT_END(bh,index) (bh==0 || (bh->next==0 && index==bh->free_index))
#define LAST_UNPARSE_ELEMENT(bh,index) (bh->next==0 && index==bh->free_index-1)

static UnparseRange text_range_to_unparse_range(Fragments f, Range R) {
  Position ue_start; /* start of screen representation of current element */
  Position ue_end;   /* end of screen representation of current element */
  int empty_range; /* is parameter R empty? */
  int has_screen_repr;
  UnparseBlockHeader bh;
  int index;

  int indentation;
  UnparseStack unparse_stack;

  UnparseRange res;
  Position current;
  int found;

  indentation=0; unparse_stack=0;
  res.start.f=f; res.start.block=0; res.start.index=0;
  res.end.f=f;   res.end.block=0;   res.end.index=0;
  empty_range=pos_eq(R.start,R.end);
  if (f==0 || f->first==0) return res;
  TRACE_POINTING_START
    printf("LOCATING UNPARSE ELEMENT RANGE. SCREEN RANGE =(%d,%d)--(%d,%d)\n", 
            R.start.row, R.start.col, R.end.row, R.end.col);
  TRACE_POINTING_END

  current.row=1; current.col=0;
  bh=f->first; index=0;

  found=0;
  while (!AT_END(bh,index) && !found) {
    ue_start=current;
    { UnparseElement unp_elem;
      unp_elem=bh->elements[index];
      switch(unp_elem.kind) {
        case TEMPLATE_STR_REF_FRAGMENT : current.col+=TEMPLATE_STR_REF(unp_elem).num_chars; break;
        case TREE_STR_REF_FRAGMENT     : current.col+=TREE_STRING_REF(unp_elem).num_chars; break;
        case TREE_CHAR_FRAGMENT        : current.col+=1; break;
        case CONSTANT_STR_FRAGMENT     : current.col+=strlen(STRING(unp_elem)); break;
        case HARD_NL_FRAGMENT          : current.row+=1; current.col=indentation; break;
        case OPTIONAL_NL_FRAGMENT      : break;
        case CONDITIONAL_NL_FRAGMENT   : break;
        case SET_INDENT_FRAGMENT       : unparse_stack=push_unparse(indentation,unparse_stack);
                                         indentation=current.col;
                                         break;
        case INDENT_UP_FRAGMENT        : unparse_stack=push_unparse(indentation,unparse_stack); indentation+=2; break;
        case INDENT_DOWN_FRAGMENT      :
            if (unparse_stack==0) { indentation=0; }
            else {
              indentation=unparse_stack->indentation;
              unparse_stack=pop_unparse(unparse_stack);
            }
            if (indentation<0) indentation=0;
            break;
        case CHAR_FRAGMENT             : current.col++; break;
        default : break;
      }
    }

    ue_end=current;
    TRACE_POINTING_START
      printf("Locating first (index=%d); ", index);
      printf("ue_start=(%d,%d); ue_end=(%d,%d);", ue_start.row, ue_start.col,
       ue_end.row, ue_end.col);


      printf("\n");
    { int contains_R, starts_at_R, ends_at_R;
      int before_R, after_R, outside_R;
      int produces_R;

      contains_R       = (pos_leq(ue_start,R.start) && pos_geq(ue_end,R.end))
;
      starts_at_R      = (pos_eq(ue_start,R.end))
;
      ends_at_R        = (pos_eq(ue_end,R.start))
;
      before_R = (pos_leq(ue_end,R.start))
;
      after_R = (pos_leq(R.end,ue_start))
;
      outside_R = ((pos_leq(ue_end,R.start))
 || (pos_leq(R.end,ue_start))
)
;
      produces_R=(
                  (empty_range && ( (pos_leq(ue_start,R.start) && pos_geq(ue_end,R.end))
 || (pos_eq(ue_start,R.end))
 || (pos_eq(ue_end,R.start))
))
                  ||
                  (!empty_range && ( !((pos_leq(ue_end,R.start))
 || (pos_leq(R.end,ue_start))
)
) )
                 );
      printf(" empty_range=");      print_bool(empty_range); printf("\n");
      printf(" contains_R=");       print_bool(contains_R); printf("\n");
      printf(" starts_at_R=");      print_bool(starts_at_R); printf("\n");
      printf(" ends_at_R=");        print_bool(ends_at_R); printf("\n");
      printf(" before_R=");         print_bool(before_R); printf("\n");
      printf(" after_R=");          print_bool(after_R); printf("\n");
      printf(" outside_R=");        print_bool(outside_R); printf("\n");
      printf(" produces R=");       print_bool(produces_R); printf("\n");
    }

    TRACE_POINTING_END

    has_screen_repr=HAS_SCREEN_REPR(bh->elements[index].kind);

    if (has_screen_repr && (
                            (empty_range && ( (pos_leq(ue_start,R.start) && pos_geq(ue_end,R.end))
 || (pos_eq(ue_start,R.end))
 || (pos_eq(ue_end,R.start))
))
                            ||
                            (!empty_range && ( !((pos_leq(ue_end,R.start))
 || (pos_leq(R.end,ue_start))
)
) )
                           )) {
      found=1;
      res.start.block=bh;
      res.start.index=index;
    }
    if (!found) if (index<bh->free_index-1) { index++; }
                else { bh=bh->next; index=0; }

  }
  if (!found) {
    res.start.block=f->last;
    res.start.index=f->last->free_index-1;
  }

  found=0; /* find an element that does not produce R */
  current=ue_start;
  while (!AT_END(bh,index) && !found) {
    ue_start=current;
    { UnparseElement unp_elem;
      unp_elem=bh->elements[index];
      switch(unp_elem.kind) {
        case TEMPLATE_STR_REF_FRAGMENT : current.col+=TEMPLATE_STR_REF(unp_elem).num_chars; break;
        case TREE_STR_REF_FRAGMENT     : current.col+=TREE_STRING_REF(unp_elem).num_chars; break;
        case TREE_CHAR_FRAGMENT        : current.col+=1; break;
        case CONSTANT_STR_FRAGMENT     : current.col+=strlen(STRING(unp_elem)); break;
        case HARD_NL_FRAGMENT          : current.row+=1; current.col=indentation; break;
        case OPTIONAL_NL_FRAGMENT      : break;
        case CONDITIONAL_NL_FRAGMENT   : break;
        case SET_INDENT_FRAGMENT       : unparse_stack=push_unparse(indentation,unparse_stack);
                                         indentation=current.col;
                                         break;
        case INDENT_UP_FRAGMENT        : unparse_stack=push_unparse(indentation,unparse_stack); indentation+=2; break;
        case INDENT_DOWN_FRAGMENT      :
            if (unparse_stack==0) { indentation=0; }
            else {
              indentation=unparse_stack->indentation;
              unparse_stack=pop_unparse(unparse_stack);
            }
            if (indentation<0) indentation=0;
            break;
        case CHAR_FRAGMENT             : current.col++; break;
        default : break;
      }
    }

    ue_end=current;
    TRACE_POINTING_START
      printf("Locating last (index=%d); ", index);
      printf("ue_start=(%d,%d); ue_end=(%d,%d);", ue_start.row, ue_start.col,
       ue_end.row, ue_end.col);


      printf("\n");
    { int contains_R, starts_at_R, ends_at_R;
      int before_R, after_R, outside_R;
      int produces_R;

      contains_R       = (pos_leq(ue_start,R.start) && pos_geq(ue_end,R.end))
;
      starts_at_R      = (pos_eq(ue_start,R.end))
;
      ends_at_R        = (pos_eq(ue_end,R.start))
;
      before_R = (pos_leq(ue_end,R.start))
;
      after_R = (pos_leq(R.end,ue_start))
;
      outside_R = ((pos_leq(ue_end,R.start))
 || (pos_leq(R.end,ue_start))
)
;
      produces_R=(
                  (empty_range && ( (pos_leq(ue_start,R.start) && pos_geq(ue_end,R.end))
 || (pos_eq(ue_start,R.end))
 || (pos_eq(ue_end,R.start))
))
                  ||
                  (!empty_range && ( !((pos_leq(ue_end,R.start))
 || (pos_leq(R.end,ue_start))
)
) )
                 );
      printf(" empty_range=");      print_bool(empty_range); printf("\n");
      printf(" contains_R=");       print_bool(contains_R); printf("\n");
      printf(" starts_at_R=");      print_bool(starts_at_R); printf("\n");
      printf(" ends_at_R=");        print_bool(ends_at_R); printf("\n");
      printf(" before_R=");         print_bool(before_R); printf("\n");
      printf(" after_R=");          print_bool(after_R); printf("\n");
      printf(" outside_R=");        print_bool(outside_R); printf("\n");
      printf(" produces R=");       print_bool(produces_R); printf("\n");
    }

    TRACE_POINTING_END

    has_screen_repr=HAS_SCREEN_REPR(bh->elements[index].kind);

    if (has_screen_repr) {
       if ((
            (empty_range && ( (pos_leq(ue_start,R.start) && pos_geq(ue_end,R.end))
 || (pos_eq(ue_start,R.end))
 || (pos_eq(ue_end,R.start))
))
            ||
            (!empty_range && ( !((pos_leq(ue_end,R.start))
 || (pos_leq(R.end,ue_start))
)
) )
           )){
         res.end.block=bh;
         res.end.index=index;
       }
       else { /* has_screen_repr && !PRODUCES_R */
         found=1;
       }
    }
    if (!found) if (index<bh->free_index-1) { index++; }
                else { bh=bh->next; index=0; }

  }


  return res;
}

#define MAX_VISIT_DEPTH 10000
#define VISIT_TABLE_PUSH(thing) if (vt_free>=MAX_VISIT_DEPTH) visit_table_overflow();\
visit_table[vt_free]=thing; vt_free++
void visit_table_overflow (void) {
  printf("ERROR: Visit Table in Textual Unparser Overflowed\n");
  exit(1);
}

static Path unparse_element_origin(UnparseElementReference unp_elem_ref) {
  Path res;
  int visit_table[MAX_VISIT_DEPTH];
  int vt_free; /* index in visit_table */
  int attr_level; /* how deep are we in unparsing attributes */
  int attr_depth;  
  Fragments fr;
  UnparseBlockHeader bh;
  int ix;

  vt_free=0; attr_level=0; attr_depth=0;

  TRACE_POINTING_START
    printf("unparse_element_origin: ");
    print_unparse_element_reference(unp_elem_ref); printf("\n");
  TRACE_POINTING_END

  fr=unp_elem_ref.f;
  bh=fr->first;
  ix=0;
  while (!(bh==unp_elem_ref.block && ix==unp_elem_ref.index)) {
     /* printf("  treating element %d",ix); */
     switch (bh->elements[ix].kind) {
       case UP_FRAGMENT :
         if(attr_level==0) {
           vt_free--;
           /* printf("  PATH POP\n"); */
         }
         else {
          attr_depth--;
          if (attr_depth==0) attr_level--;
         }
        break;
       case CHILD_FRAGMENT : 
         if (attr_level==0) {
           VISIT_TABLE_PUSH(INT(bh->elements[ix]));
           /* printf("  PATH PUSH CHILD %d\n", INT(bh->elements[ix]) ); */
         }
         else {
           attr_depth++;
         }
         break;
       case ENTER_ATTR_FRAGMENT : attr_level++; attr_depth++; break;
       default : break;
     }
     /* printf(" vt_free=%d; attr_level=%d; attr_depth=%d\n", vt_free, attr_level, attr_depth);
      */

     if (ix<bh->free_index-1) { ix++; }
     else {
       bh=bh->next;
       ix=0;
     }

  }
  TRACE_POINTING_START
    { int vt_ix;
      printf("RESULT of unparse_element_origin: /");
      for(vt_ix=0; vt_ix<vt_free; vt_ix++) {
        printf("%d/", visit_table[vt_ix]);
      }
      printf("\n");
    }
  TRACE_POINTING_END
  res=0;
  { int vt_ix;
    for(vt_ix=vt_free-1; vt_ix>=0; vt_ix--) {
      res=path_extend(res,visit_table[vt_ix]);
    }
  }
  return res;

}

static PathPair unparse_range_to_path(UnparseRange u_range) {
  PathPair res;
  res.first=unparse_element_origin(u_range.start);
  if (unparse_ref_equal(u_range.start,u_range.end)) {
    res.second=res.first;
  }
  else {
    res.second=unparse_element_origin(u_range.end);
  }
  return res;
}

PathPair fragments_range_to_path(Fragments fr, Range screen_range) {
  UnparseRange ue_range;
  PathPair res;
  TRACE_POINTING_START
    printf("fragments_range_to_path:(%d,%d)--(%d,%d)\n",
             screen_range.start.row,screen_range.start.col,
             screen_range.end.row,  screen_range.end.col
    );
  TRACE_POINTING_END
  ue_range=text_range_to_unparse_range(fr,screen_range);
  TRACE_POINTING_START
    if (ue_range.start.f==0) { printf("ue_range.start=<null>"); }
    else {
      if (ue_range.start.block==0) printf("block==<null>");
      else {
        printf("ue_range.start= index=%d",ue_range.start.index);
        print_unparse_element_kind(ue_range.start.block->elements[ue_range.start.index].kind);
      }
    }
    printf("\n");
    if (ue_range.end.f==0) { printf("ue_range.end=<null>"); }
    else {
      if (ue_range.end.block==0) printf("block==<null>");
      else {
        printf("ue_range.end= index=%d",ue_range.end.index);
        print_unparse_element_kind(ue_range.end.block->elements[ue_range.end.index].kind);
      }
    }
    printf("\n");
  TRACE_POINTING_END
  res=unparse_range_to_path(ue_range);
  return res;
} 

Range path_to_screen_range(Fragments f, Path P) {
  UnparseBlockHeader bh;
  int index;

  int visit_table[MAX_VISIT_DEPTH];
  int vt_free; /* index in visit_table */
  Path p;
  int p_edge; /* edge of p we want to traverse; */
  int watch_position; /* the position in the table that we are watching */
  Position current;
  int indentation;
  UnparseStack unparse_stack;
  int attr_level, attr_depth;
  Range res;
  int found;

  bh=0;
  if (f!=0) bh=f->first;
  index=0;


  vt_free=0;
  p=P;
  watch_position=0;

  current.row=1; current.col=0;
  indentation=0;
  unparse_stack=0;
  attr_level=0;
  attr_depth=0;

  res.start.row=1; res.start.col=0;
  res.end.row  =1; res.end.col  =0;
  if (f==0 || f->first==0) return res;

  TRACE_POINTING_START
    printf("path_to_screen_range: ");
    path_print(P); printf("\n");
  TRACE_POINTING_END

  if (P!=0) {
    bh=f->first; index=0;

    p_edge=p->childnumber;
    found=0;
    while (!found && !AT_END(bh,index)
) {
      { UnparseElement unp_elem;
        unp_elem=bh->elements[index];
        switch(unp_elem.kind) {
          case TEMPLATE_STR_REF_FRAGMENT : current.col+=TEMPLATE_STR_REF(unp_elem).num_chars; break;
          case TREE_STR_REF_FRAGMENT     : current.col+=TREE_STRING_REF(unp_elem).num_chars; break;
          case TREE_CHAR_FRAGMENT        : current.col+=1; break;
          case CONSTANT_STR_FRAGMENT     : current.col+=strlen(STRING(unp_elem)); break;
          case HARD_NL_FRAGMENT          : current.row+=1; current.col=indentation; break;
          case OPTIONAL_NL_FRAGMENT      : break;
          case CONDITIONAL_NL_FRAGMENT   : break;
          case SET_INDENT_FRAGMENT       : unparse_stack=push_unparse(indentation,unparse_stack);
                                           indentation=current.col;
                                           break;
          case INDENT_UP_FRAGMENT        : unparse_stack=push_unparse(indentation,unparse_stack); indentation+=2; break;
          case INDENT_DOWN_FRAGMENT      :
              if (unparse_stack==0) { indentation=0; }
              else {
                indentation=unparse_stack->indentation;
                unparse_stack=pop_unparse(unparse_stack);
              }
              if (indentation<0) indentation=0;
              break;
          case CHAR_FRAGMENT             : current.col++; break;
          default : break;
        }
      }

      switch (bh->elements[index].kind) {
        case UP_FRAGMENT :
          if(attr_level==0) {
            vt_free--;
          }
          else {
           attr_depth--;
           if (attr_depth==0) attr_level--;
          }
         break;
        case CHILD_FRAGMENT :
          if (attr_level==0) { int edge;
                               edge=INT(bh->elements[index]);
                               if (vt_free==watch_position && p_edge==edge) {
                                 { watch_position++;
                                   p=p->next;
                                   if (p!=0) { p_edge=p->childnumber; }
                                   else { found=1; }
                                 }

                               }
                               { VISIT_TABLE_PUSH(INT(bh->elements[index]));
                               }
                             }

          else {
            attr_depth++;
          }
          break;
        case ENTER_ATTR_FRAGMENT : attr_level++; attr_depth++; break;
        default : break;
      }

      if (found) {
        res.start.row=current.row;
        res.start.col=current.col;
      }
      if (!found) if (index<bh->free_index-1) { index++; }
                  else { bh=bh->next; index=0; }

    }

  }
  found=0;
  while (!found && !AT_END(bh,index)
) {
    { UnparseElement unp_elem;
      unp_elem=bh->elements[index];
      switch(unp_elem.kind) {
        case TEMPLATE_STR_REF_FRAGMENT : current.col+=TEMPLATE_STR_REF(unp_elem).num_chars; break;
        case TREE_STR_REF_FRAGMENT     : current.col+=TREE_STRING_REF(unp_elem).num_chars; break;
        case TREE_CHAR_FRAGMENT        : current.col+=1; break;
        case CONSTANT_STR_FRAGMENT     : current.col+=strlen(STRING(unp_elem)); break;
        case HARD_NL_FRAGMENT          : current.row+=1; current.col=indentation; break;
        case OPTIONAL_NL_FRAGMENT      : break;
        case CONDITIONAL_NL_FRAGMENT   : break;
        case SET_INDENT_FRAGMENT       : unparse_stack=push_unparse(indentation,unparse_stack);
                                         indentation=current.col;
                                         break;
        case INDENT_UP_FRAGMENT        : unparse_stack=push_unparse(indentation,unparse_stack); indentation+=2; break;
        case INDENT_DOWN_FRAGMENT      :
            if (unparse_stack==0) { indentation=0; }
            else {
              indentation=unparse_stack->indentation;
              unparse_stack=pop_unparse(unparse_stack);
            }
            if (indentation<0) indentation=0;
            break;
        case CHAR_FRAGMENT             : current.col++; break;
        default : break;
      }
    }

    switch (bh->elements[index].kind) {
      case UP_FRAGMENT :
        if(attr_level==0) {
          vt_free--;
          if ((  (P==0 && LAST_UNPARSE_ELEMENT(bh,index))
              || (P!=0 && (vt_free==watch_position))
              )
) found=1;
        }
        else {
         attr_depth--;
         if (attr_depth==0) attr_level--;
        }
       break;
      case CHILD_FRAGMENT :
        if (attr_level==0) { VISIT_TABLE_PUSH(INT(bh->elements[index]));
                           }
        else {
          attr_depth++;
        }
        break;
      case ENTER_ATTR_FRAGMENT : attr_level++; attr_depth++; break;
      default : break;
    }
    if (found) {
      res.end.row=current.row;
      res.end.col=current.col;
    }
    if (!found) if (index<bh->free_index-1) { index++; }
                else { bh=bh->next; index=0; }

  }


  TRACE_POINTING_START
    printf("  --> (%d,%d)-(%d,%d)\n",
           res.start.row, res.start.col,
           res.end.row, res.end.col
    );
  TRACE_POINTING_END

  return res;
}


