#ifndef H_FRAGMENTS
#define H_FRAGMENTS
typedef enum {
  NO_FRAGMENT=0,
  CHILD_FRAGMENT,
  UP_FRAGMENT,
  INDENT_UP_FRAGMENT,
  INDENT_DOWN_FRAGMENT,
  OPEN_GROUP_FRAGMENT,
  CLOSE_GROUP_FRAGMENT,
  ORIGIN_FRAGMENT,
  PUSH_STYLE_FRAGMENT,
  POP_STYLE_FRAGMENT,
  SET_INDENT_FRAGMENT,
  ENTER_ATTR_FRAGMENT,
  TEMPLATE_STR_REF_FRAGMENT,
  TREE_STR_REF_FRAGMENT,
  TREE_CHAR_FRAGMENT,
  CONSTANT_STR_FRAGMENT,
  HARD_NL_FRAGMENT,
  OPTIONAL_NL_FRAGMENT,
  CONDITIONAL_NL_FRAGMENT,
  CHAR_FRAGMENT
} UnparseElementKind;

typedef struct template_string_ref {
  int template_index; /* index in an template table */
  short int offset;  /* offset of first character in string */
  short int num_chars; /* number of characters in string */
} TemplateStringRef;

typedef struct tree_string_ref {
  lrc_term term;
  short int offset;  /* offset of first character in string */
  short int num_chars; /* number of characters in string */
} TreeStringRef;

typedef struct unparse_element {
  UnparseElementKind kind;
  union {
    TemplateStringRef tmpl_sr;
    TreeStringRef     tree_sr;
    lrc_term          term;
    lrc_string        string;
    char              ch;
    int               i;
  } element_data;
  
} UnparseElement;



typedef UnparseElement * UnparseBlock;

typedef struct unparse_block_header {
   struct unparse_block_header * next;
   struct unparse_block_header * prev;
   int num_elements;
   int free_index;
   UnparseBlock elements;
} *UnparseBlockHeader;

typedef struct fragments {
   UnparseBlockHeader first;
   UnparseBlockHeader last;
} * Fragments;

typedef struct unparse_stack_element {
  struct unparse_stack_element * below;
  int indentation;
  lrc_string indentation_string;
} UnparseStackElement, * UnparseStack;


typedef struct unparse_element_ref {
  Fragments f;
  UnparseBlockHeader block;
  int index;
} UnparseElementReference;

typedef struct unparse_range {
  UnparseElementReference start;
  UnparseElementReference end;
} UnparseRange;

#define HAS_SCREEN_REPR(k) ((k)>ENTER_ATTR_FRAGMENT)

#define TEMPLATE_STR_REF(e) e.element_data.tmpl_sr
#define TREE_STRING_REF(e) e.element_data.tree_sr
#define STRING(e) e.element_data.string
#define CHAR(e) e.element_data.ch
#define INT(e) e.element_data.i
#define TERM(e) e.element_data.term

#define MIN_UNPARSE_BLOCK_SIZE 1000

#define FREE_INDEX(f) f->last->free_index
#define FREE_ELEMENT(f) f->last->elements[FREE_INDEX(f)]
#define FREE_DATA(f) FREE_ELEMENT(f).element_data
#define BUMP(f)  FREE_INDEX(f)=FREE_INDEX(f)+1

void print_unparse_element(UnparseElement e, unparse_commands_tp * commands);

void fragments_print_block_info(Fragments f);


UnparseStack push_unparse(int indentation, UnparseStack us);
UnparseStack pop_unparse(UnparseStack us);


Fragments create_fragments(void);
void fragments_delete_elements(Fragments fr);

void fragments_all_elements(Fragments f,  void action());

void fr_add_template_substring(int template_index, int char_offset, int length, Fragments f);
void fr_add_tree_substring(lrc_term t, int char_offset, int length, Fragments f);
void fr_add_tree_char(lrc_term t, Fragments f);
void fr_add_constant_string(char * str, Fragments f);
void fr_add_indent_up(Fragments f);
void fr_add_indent_down(Fragments f);
void fr_add_set_indent(Fragments f);
void fr_add_line_break(Fragments f);
void fr_add_optional_line_break(Fragments f);
void fr_add_consistent_line_break(Fragments f);
void fr_add_open_group(Fragments f);
void fr_add_close_group(Fragments f);
void fr_add_constant_char(char c, Fragments f);
void fr_add_push_style(char * style_name, int style_name_length, Fragments f);
void fr_add_pop_style(Fragments f);
void fr_add_tree_str(lrc_term t, Fragments f);
void fr_add_origin(lrc_term t, Fragments f);
void fr_add_child(int n, Fragments f);
void fr_add_up(Fragments f);
void fr_add_enter_attr(Fragments f);

PathPair fragments_range_to_path(Fragments fr, Range screen_range);

Range path_to_screen_range(Fragments f, Path P);


#endif

