#include <stdio.h>
#include "maps.h"
#include "ints.h"
#include "bools.h"
#include "strs.h"
#include "term.h"
#include "prims.h"
#include "grammar_tables.h"
#include "map_operations.h"

#define map_ib_new(def) lrc_empty_map(lrc_int_equal_term,lrc_int_compare_term,lrc_bool_equal_term,lrc_bool_compare_term,C_Bool(def))
#define map_ii_new(def) lrc_empty_map(lrc_int_equal_term,lrc_int_compare_term,lrc_int_equal_term,lrc_int_compare_term,C_Int(def))
#define map_is_new(def) lrc_empty_map(lrc_int_equal_term,lrc_int_compare_term,lrc_string_equal_term,lrc_string_compare_term,C_Str(def))
#define map_si_new(def) lrc_empty_map(lrc_string_equal_term,lrc_string_compare_term,lrc_int_equal_term,lrc_int_compare_term,C_Int(def))
#define map_tb_new(def) lrc_empty_map(lrc_term_equal,LrcCompareTrees,lrc_bool_equal_term,lrc_bool_compare_term,C_Bool(def))
#define map_ti_new(def) lrc_empty_map(lrc_term_equal,LrcCompareTrees,lrc_int_equal_term,lrc_int_compare_term,C_Int(def))

lrc_bool map_ib_apply(lrc_map map, int int_key) {
  return LRC_VALUE_lrc_bool(lrc_map_application(C_Int(int_key),map));
}

int map_ii_apply(lrc_map map, int int_key) {
  return LRC_VALUE_int(lrc_map_application(C_Int(int_key),map));
}

lrc_string map_is_apply(lrc_map map, int int_key) {
  return LRC_VALUE_lrc_string(lrc_map_application(C_Int(int_key),map));
}

int map_si_apply(lrc_map map, lrc_string str_key) {
  return LRC_VALUE_int(lrc_map_application(C_Str(str_key),map));
}

lrc_bool map_tb_apply(lrc_map map, lrc_term trm_key) {
  return LRC_VALUE_lrc_bool(lrc_map_application(trm_key,map));
}

int map_ti_apply(lrc_map map, lrc_term trm_key) {
  int res;
  res=LRC_VALUE_int(lrc_map_application(trm_key,map));
  return res;
}

lrc_map map_ib_set(lrc_map map, int int_key, lrc_bool bool_val) {
   lrc_map result;
   result=lrc_map_single_update(C_Int(int_key), C_Bool(bool_val), map);
   return result;
}

lrc_map map_ii_set(lrc_map map, int int_key, int int_val) {
   lrc_map result;
   result=lrc_map_single_update(C_Int(int_key), C_Int(int_val), map);
   return result;
}

lrc_map map_is_set(lrc_map map, int int_key, lrc_string string_val) {
   lrc_map result;
   result=lrc_map_single_update(C_Int(int_key), C_Str(string_val), map);
   return result;
}

lrc_map map_si_set(lrc_map map, lrc_string string_key, int int_val) {
   lrc_map result;
   result=lrc_map_single_update(C_Str(string_key), C_Int(int_val), map);
   return result;
}

lrc_map map_ti_set(lrc_map map, lrc_term term_key, int int_val) {
   lrc_map result;
   result=lrc_map_single_update(term_key, C_Int(int_val), map);
   return result;
}

lrc_map map_tb_set(lrc_map map, lrc_term term_key, lrc_bool bool_val) {
   lrc_map result;
   result=lrc_map_single_update(term_key, C_Bool(bool_val), map);
   return result;
}

lrc_map lrc_map_default(void) {
  lrc_map res;
  printf("DANGER: Constructing a default MAP!!!\n");
  res=map_ii_new(0);
  return res;
}


static lrc_term avl_keys(AVL_tp a, int bin_prod, lrc_term keys_in) {
 lrc_term right_out;
 if (a==0) return keys_in;
 right_out=avl_keys(a->right, bin_prod, keys_in);
 return avl_keys( a->left, bin_prod, lrc_term2(bin_prod,a->key,right_out));
}

static lrc_term avl_values(AVL_tp a, int bin_prod, lrc_term values_in) {
 lrc_term right_out;
 if (a==0) return values_in;
 right_out=avl_values(a->right, bin_prod, values_in);
 return avl_values( a->left, bin_prod, lrc_term2(bin_prod,a->value,right_out));
}

static lrc_term avl_map_values(AVL_tp a, lrc_term f(), int bin_prod, lrc_term list_in) {
 lrc_term right_out;
 if (a==0) return list_in;
 right_out=avl_map_values(a->right, f, bin_prod, list_in);
 return avl_map_values( a->left, f, bin_prod, lrc_term2(bin_prod,f(a->value),right_out));
}

lrc_term map_keys(lrc_map map, int bin_prod, int null_prod) {
  return avl_keys(map.avl, bin_prod, lrc_term0(null_prod));
}

lrc_term map_values(lrc_map map, int bin_prod, int null_prod) {
  return avl_values(map.avl, bin_prod, lrc_term0(null_prod));
}

lrc_term map_map_values(lrc_map map, lrc_term f(), int bin_prod, int null_prod) {
  return avl_map_values(map.avl, f, bin_prod, lrc_term0(null_prod));
}



