/*
** maps.c
** march 1994 by Maarten Pennings
** Library functions by Matthijs Kuiper
*/

#include <stdio.h>
#include "maps.h"

static lrc_term_ix index_map(lrc_map m) {
  return ( lrc_term_hashix(m.default_value) 
         + avl_hash(m.avl)
         + T__Map
         ) % lrc_term_tabelsize;
}

static lrc_bool eq_map(lrc_term t,lrc_ptr b) {
  return LRC_TAG(t)==FIELD(b,0,int) && 
    lrc_map_equal(LRC_VALUE(t,lrc_map),FIELD(b,sizeof(lrc_ptr),lrc_map));
}

lrc_term C_Map(lrc_map m) {
  struct {lrc_ptr tag;lrc_map m;} buf;
  buf.tag=(lrc_ptr)T__Map; buf.m=m;
  return lrc_shared_prim(index_map(m),eq_map,sizeof(buf),&buf);
}

void map_forall(lrc_map m, void action()) {
  avl_forall(m.avl,action);
}

void map_forall_keys(lrc_map m, void action()) {
  avl_forall_keys(m.avl,action);
}

int map_num_keys(lrc_map m) {
  return avl_num_elements(m.avl);
}

/*********************\
** Library functions **
\*********************/


int lrc_map_compare(lrc_map a, lrc_map b) {
 /* a.equal_key==b.equal_key && a.equal_value==b.equal_value) */
 int default_cmp=a.compare_value(a.default_value,b.default_value);
 if (default_cmp!=lrc_equal) return default_cmp;
 return avl_compare(a.avl,b.avl
                   ,a.equal_key,a.compare_key
                   ,a.equal_value,a.compare_value);
}

int lrc_map_compare_term(lrc_term a, lrc_term b) {
  return lrc_map_compare(LRC_VALUE(a,lrc_map),LRC_VALUE(b,lrc_map));
}

lrc_bool lrc_map_equal(lrc_map a, lrc_map b) {
 /* a.equal_key==b.equal_key && a.equal_value==b.equal_value) */
 if (!a.equal_value(a.default_value,b.default_value)) return lrc_false;
 return avl_equal(a.avl,b.avl,a.equal_key,a.equal_value);
}

lrc_bool lrc_map_equal_term(lrc_term a, lrc_term b) {
  return lrc_map_equal(LRC_VALUE(a,lrc_map),LRC_VALUE(b,lrc_map));
}

void lrc_map_print(lrc_map m, void print_key(), void print_value()) {
  printf("["); print_value(m.default_value); printf(";");
  avl_print(m.avl,print_key,print_value);
  printf("]");
  return;
}

/*
#define AssignMap(dest,src) \
   dest.default_value=src.default_value;\
   dest.avl=src.avl;\
   dest.equal_key=src.equal_key;\
   dest.compare_key=src.compare_key;\
   dest.equal_value=src.equal_value;\
   dest.compare_value=src.compare_value;
*/

lrc_term lrc_map_application(lrc_term k, lrc_map m) {
   lrc_term res;

   if (avl_lookup(&res,k,m.avl,m.compare_key)) return res;
   return m.default_value;
}

lrc_bool lrc_map_in(lrc_term k, lrc_map m) {
   lrc_term res;

   if (avl_lookup(&res,k,m.avl,m.compare_key)) return lrc_true;
   return lrc_false;
}

lrc_map lrc_map_single_update(lrc_term key, lrc_term value, lrc_map m) {
  lrc_map res=m;
  /* AssignMap(res,m); */
  res.avl=avl_insert(key,value,m.avl,m.compare_key);
  return res;
}

lrc_map lrc_map_update(lrc_map m0, lrc_map m1) {
  /* return   m0 overwritten_with m1   */
  lrc_map res=m0;
  /* AssignMap(res,m0); */ 
  res.avl=avl_update(m0.avl,m1.avl,m0.compare_key);
  return res;
}

lrc_map lrc_empty_map(
      equal_function_tp *equal_key,
      compare_function_tp *compare_key,
      equal_function_tp *equal_value,
      compare_function_tp *compare_value,
      lrc_term def
   ) {
   lrc_map res;
   res.default_value=def;
   res.equal_key=equal_key;
   res.compare_key=compare_key;
   res.equal_value=equal_value;
   res.compare_value=compare_value;
   res.avl=avl_empty();
   return res;
}

/* add all elements from value_list to m;
 * elements are key-value pairs
 */
lrc_map lrc_map_init(lrc_map m, lrc_term key_value_list, int end_of_list_tag) {
   lrc_term kvl;
   lrc_map res=m;
   /* AssignMap(res,m); */
   kvl=key_value_list;
   while (LRC_TAG(kvl)!=end_of_list_tag) {
     res = lrc_map_single_update(
                      LRC_CHILD(LRC_CHILD(kvl,1),1),
                      LRC_CHILD(LRC_CHILD(kvl,1),2),
		      res);
     kvl = LRC_CHILD(kvl,2);
   }
   return res;
}

/* Marking */
void mark_map(lrc_term t) {
  lrc_map m=LRC_VALUE(t,lrc_map);
  mark_term(m.default_value);
  mark_avl(m.avl);
}

