/*
** memo.c
** february 1994 Maarten Pennings
** (implemented the full memoing version)
**
** march    1996 Joao Saraiva
** Delta and Generations strategies
**
** Generated from memo_gen.nw
**
**
*/

#include <stdlib.h>
#include <memory.h>
#include <stdio.h>
#include <math.h>
#include "lrcmemory.h"
#include "prims.h"
#include "memo.h"

#define EMPTY          0
#define ALLOC_CELL(n)  lrc_malloc(n)

#define NOMEMOING   0
#define FULLMEMOING 1
#define DELTACACHE  2
#define GENERATIONS 3

typedef lrc_cell cache_entry;
typedef lrc_cell cache;

typedef struct gen
{ unsigned int  gen_age;
  cache_entry  *gen;
  struct gen   *next;
} *list_gen;


#define GEN_AGE(g)        g->gen_age
#define GEN(g)            g->gen
#define NEXT_GEN(g)       g->next

typedef struct
{ unsigned int numb_entries;
  lrc_cell list;
} bucket_info;


static bucket_info table_bucket[lrc_memo_tablesize];
#define NBE(e)     table_bucket[e].numb_entries
#define LIST(e)    table_bucket[e].list


static int lrc_memo_full_memoing (lrc_visfun f,int ipsize,lrc_packet ip,
                                               int opsize,lrc_packet op);

static int replac_func_bucket_LRU(int ind_bucket);

static int       memo_strateg = 1;
static lrc_bool  memo_tree_size = 0;
static int stats_min_tree_size_to_cache = 0;
static int stats_numb_trees_below_size;

static int  number_G;
static int  number_L;

static list_gen     l_generations;
static list_gen     l_younger_gen;
static cache       *oldest_gen;
static cache_entry *table;


static int stats_num_allocs;
static int stats_num_hits;
static int stats_size_allocs;
static int stats_memory_used;
static int stats_num_entries;

static int stats_num_generations;
static lrc_bool stats_copying_to_old_gen;
static int stats_one_more_year;

#ifdef COMP_STATISTICS
#ifdef COMP_STATISTICS
static int stats_num_hash_op;
static int stats_num_copy_op;
static int stats_num_bytes_copied;
static int stats_num_copy_op_old_gen;
static int stats_num_bytes_copied_old_gen;
static int stats_hit_in_old;
static int stats_hit_in_all;
static int stats_hit_in_new;

static int stats_total_num_hash_op;
static int stats_total_num_copy_op;
static int stats_total_num_bytes_copied;
static int stats_total_num_copy_op_old_gen;
static int stats_total_num_bytes_copied_old_gen;
static int stats_total_num_hits;
static int stats_total_num_allocs;
stats_num_replac;
#endif

#endif


void lrc_memo_strategy_config (int memoing_strategy, int g, int l)
{

  memo_strateg = memoing_strategy;
  if ( memo_strateg == GENERATIONS && g < 2 )
    number_G = 2;
  else number_G = g;
  number_L = l;
  #ifdef WRITE_RUNTIME_STRATEGY
    switch (memo_strateg)
      { case NOMEMOING: fprintf(stdout," Evaluation without memoization\n");
        break;
        case FULLMEMOING: fprintf(stdout," Evaluation using Full Memoization\n");
        break;
      case DELTACACHE: fprintf(stdout," Evaluation using the Delta Cache\n");
        break;
      case GENERATIONS: fprintf(stdout," Evaluation using Generations\n");
        fprintf(stdout," number of eval. for a collection: %d ",number_G);
        fprintf(stdout," number of cache instances: %d \n",number_L);
  #ifdef OLDEST_NEW_YOUNGEST
        fprintf(stdout," Lookup strategy: oldest - new - youngest\n");
  #endif
  #ifdef YOUNGEST_OLDEST_NEW
        fprintf(stdout," Lookup strategy: youngest - oldest - new");
  #endif
  #ifdef YOUNGEST_NEW_OLDEST
        fprintf(stdout," Lookup strategy: youngest - new - oldest\n");
  #endif
  #ifdef OLDEST_ALLYOUNGER_NEW
        fprintf(stdout," Lookup strategy: oldest - younger - new\n");
  #endif
  #ifdef ALLYOUNGER_OLDEST_NEW
        fprintf(stdout," Lookup strategy: younger - oldest - new\n");
  #endif
  #ifdef NEW_YOUNGEST_OLDEST
        fprintf(stdout," Lookup strategy: new - youngest - oldest\n");
  #endif
  #ifdef NEW_OLDEST_YOUNGEST
        fprintf(stdout," Lookup strategy: new - oldest - youngest\n");
  #endif
      }
  #endif

}

void lrc_memo_tree_size_config (int tree_size, int size_of_subtree)
{
  fprintf(stdout," Tree S: %d %d \n",tree_size,size_of_subtree);

  memo_tree_size = tree_size;
  stats_min_tree_size_to_cache = size_of_subtree;
}

static cache * create_gen()
{ cache_entry * c;
  int i;

  c = (cache_entry *)ALLOC_CELL(sizeof(cache_entry) * lrc_memo_tablesize);
  for ( i = 0 ; i < lrc_memo_tablesize ; ++i )
     *(c+i) = EMPTY;
  return c;
}

void lrc_memo_table_old_clear(void)
{ lrc_memo_index i;
  for ( i=0 ; i < lrc_memo_tablesize ; i++ )
    { NBE(i) = 0;
      LIST(i) = EMPTY;
    }
}


void lrc_memo_init()
{
  l_generations = EMPTY;
  l_younger_gen = EMPTY;

  lrc_memo_table_old_clear();

  stats_one_more_year = 0;
  stats_copying_to_old_gen = lrc_false;

  stats_num_allocs = 0;
  stats_num_hits = 0;
  stats_size_allocs = 0;
  stats_memory_used = lrc_memo_tablesize * sizeof(lrc_ptr);
  stats_num_entries = 0;
  #ifdef COMP_STATISTICS
  stats_num_hash_op = 0;
  stats_num_copy_op = 0;
  stats_num_bytes_copied = 0;
  stats_num_copy_op_old_gen = 0;
  stats_num_bytes_copied_old_gen = 0;
  stats_hit_in_old = 0;
  stats_hit_in_all = 0;
  stats_hit_in_new = 0;
  #endif

  #ifdef COMP_STATISTICS
  stats_total_num_hash_op = 0;
  stats_total_num_copy_op = 0;
  stats_total_num_bytes_copied = 0;
  stats_total_num_copy_op_old_gen = 0;
  stats_total_num_bytes_copied_old_gen = 0;
  stats_total_num_hits = 0;
  stats_total_num_allocs = 0;
  stats_num_replac = 0;
  #endif


}

void lrc_memo_start_AE()
{ int i;
  list_gen new_gen;

  stats_one_more_year++;
  stats_num_hits = 0;
  stats_num_allocs = 0;
  #ifdef COMP_STATISTICS
  stats_num_hash_op = 0;
  stats_num_copy_op = 0;
  stats_num_bytes_copied = 0;
  stats_num_copy_op_old_gen = 0;
  stats_num_bytes_copied_old_gen = 0;
  stats_hit_in_old = 0;
  stats_hit_in_all = 0;
  stats_hit_in_new = 0;
  #endif


  switch ( memo_strateg )
    {  case FULLMEMOING:
         if ( l_generations == EMPTY )
           { l_generations = (list_gen)lrc_malloc(sizeof(struct gen));
             GEN_AGE(l_generations) = 0;
             GEN(l_generations) = create_gen();
             NEXT_GEN(l_generations) = EMPTY;
           }
         break;
       case DELTACACHE:
         new_gen = (list_gen)lrc_malloc(sizeof(struct gen));
         GEN(new_gen) = create_gen();
         NEXT_GEN(new_gen) = l_generations;
         l_generations = new_gen;

         if ( NEXT_GEN(l_generations) != EMPTY )
           NEXT_GEN(NEXT_GEN(l_generations)) = EMPTY;
         break;
       case GENERATIONS:
         if ( stats_one_more_year % number_G == 0 )
           { fprintf(stdout," Performing one Collection!!\n");
             stats_copying_to_old_gen = lrc_true;
           }
         else
           stats_copying_to_old_gen = lrc_false;

         new_gen = (list_gen)lrc_malloc(sizeof(struct gen));
         GEN_AGE(new_gen) = 0;
         GEN(new_gen) = create_gen();
         NEXT_GEN(new_gen) = l_generations;
         l_younger_gen = l_generations;
         l_generations = new_gen;

         for ( new_gen = l_younger_gen , i = 1 ;
              (new_gen != EMPTY) && (i < number_L) ;
              new_gen = NEXT_GEN(new_gen) , i++)
           ;
         if ( new_gen != EMPTY )
           NEXT_GEN(new_gen) = EMPTY;
   }
}

void lrc_memo_end_AE()
{
  #ifdef COMP_STATISTICS
    stats_total_num_hash_op += stats_num_hash_op;
    stats_total_num_copy_op += stats_num_copy_op;
    stats_total_num_bytes_copied += stats_num_bytes_copied;
    stats_total_num_copy_op_old_gen +=  stats_num_copy_op_old_gen;
    stats_total_num_bytes_copied_old_gen += stats_num_bytes_copied_old_gen;
    stats_total_num_hits += stats_num_hits;
    stats_total_num_allocs += stats_num_allocs;
    fprintf(stdout," TOTALS -- Misses: %d   Hits: %d   Lookups: %d  Copy Op.: %d  Bytes Cpoied: %d \n",
      stats_total_num_allocs,stats_total_num_hits,stats_total_num_hash_op,
      stats_total_num_copy_op,stats_total_num_bytes_copied);
    fprintf(stdout,"      Old Gen: Copy Oper: %d  Bytes Copies: %d \n",
      stats_total_num_copy_op_old_gen,stats_total_num_bytes_copied_old_gen);
  #endif

}

void copy_to_old()
{
  oldest_gen = GEN(l_generations);
}

/*
**   Every entry in a lrc_cell is a pointer.
**
**      0      1       2      3     4    5
**   +------+------+------+-----+------+------+- ... -+------+------+- ... -+------+
**   | next | hits | birth| mem | func | ptrs in inputpacket | ptrs in outputpacket|
**   +------+------+------+-----+------+------+- ... -+------+------+- ... -+------+
**
*/
#define ADMIN_SIZE       CELL_ADMIN_SIZE
#define CELL_SIZE(si,so) ((ADMIN_SIZE+1)*sizeof(lrc_ptr)+(si)+(so))

#define ADMIN(c,n,tp)    (*(((tp*)(c))+(n)))
#define NEXT(c)          ADMIN(c,0,lrc_cell)
#define HITS(c)          ADMIN(c,1,int)
#define BIRTH(c)         ADMIN(c,2,int)
#define MEM(c)           ADMIN(c,3,int)
#define FUNC(c)          ADMIN(c,4,lrc_visfun)

#define BUFFER(c,n)      ((c)+(n+ADMIN_SIZE))
#define IP(c)            BUFFER(c,1)
#define OP(c,n)          BUFFER(c,1+n)

#define LRC_ABS(i)              ((i)<0?-(i):(i))
#define PTR_TO_POS_INT(p)       (LRC_ABS((int)p))
#define HASH(p)                 (PTR_TO_POS_INT(p) % lrc_memo_tablesize)
#define REHASH_IX_PTR(ix,ptr)   ((2*(ix)+PTR_TO_POS_INT(ptr)) % lrc_memo_tablesize)

#define EQ(f,ip,size,c)   (f==FUNC(c) && memcmp(IP(c),ip,size)==0)


static int lrc_memo_delta_cache (lrc_visfun f,int ipsize,lrc_packet ip,
                                              int opsize,lrc_packet op)
{ int goon;
  lrc_cell c,c1;
  int ipnum=ipsize/sizeof(lrc_ptr);
  int num_func_remv;
  lrc_cell new_cell;           /* Variable to store the function call */
  list_gen g;
  int mem,ag,t;
  lrc_bool found_in_new_cache = lrc_false;
  lrc_bool copied_old_gen = lrc_false;
  lrc_bool found_in_oldest = lrc_false;
  lrc_bool found_in_all    = lrc_false;

  lrc_memo_index i =  HASH(f);
  { int n;
    for(n=0; n<ipnum; n++) i=REHASH_IX_PTR(i,ip[n]);
  }


  { lrc_bool found = lrc_false;
    for (g = l_generations ; !found && g != EMPTY  ; g = NEXT_GEN(g) )
      {
        c = (lrc_cell) (*(GEN(g) + i));
  #ifdef COMP_STATISTICS
        while ( c != EMPTY && !found )
          { if ( EQ(f,ip,ipsize,c) )
              { found = lrc_true;
                stats_num_hash_op++;
              }
            else
              c=NEXT(c);
          }
  #else
        while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
          c = NEXT(c);
        found = (c != EMPTY);
  #endif
      }
  }

  if ( c == EMPTY )
    { /************ not in cache, compute & store it */
      lrc_cell new_cell;
      f(ip,op);                           /* Call the function */
      new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
      c = new_cell;
      *c++ = *(GEN(l_generations) + i);            /* NEXT */
      *c++ = (lrc_cell) 0;                         /* HITS */
      *c++ = (lrc_cell) stats_num_allocs;          /* BIRTH */
      *c++ = (lrc_cell) CELL_SIZE(ipsize,opsize);  /* MEM */
      *(GEN(l_generations) + i) = new_cell;
      *c++ = (lrc_cell)f;                          /* FUNC */
      memcpy(c,ip,ipsize);
      c += ipnum;
      memcpy(c,op,opsize);
      stats_num_allocs++;
      stats_size_allocs += CELL_SIZE(ipsize,opsize);
      stats_memory_used += CELL_SIZE(ipsize,opsize);
      stats_num_entries++;
      #ifdef FUNC_HIST
      lrc_memo_f_set_calls(lrc_vpr_index(f));
      #endif

      return 0;



    }
  else
    { /****************************** in cache, retrieve it */
      stats_num_hits++;
      HITS(c)++;
      /* change the birth-time to reflect hit */
      BIRTH(c)=stats_num_allocs;
      memcpy(op,OP(c,ipnum),opsize);
      if ( ag > 1 )
      { new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
        c1 = new_cell;
        *c1++ = *(GEN(l_generations) + i);             /* NEXT */
        *c1++ = (lrc_cell) 0;                         /* HITS */
        *c1++ = (lrc_cell) stats_num_allocs;          /* BIRTH */
        memcpy(c1,c+3,ipsize + opsize + 2 * sizeof(lrc_cell));
        *(GEN(l_generations) + i) = new_cell;
        #ifdef COMP_STATISTICS
        stats_num_copy_op++;
        stats_num_bytes_copied += MEM(new_cell);
        #endif

      }



      return HITS(c);

    }
}

static int lrc_memo_generations (lrc_visfun f,int ipsize,lrc_packet ip,
                                              int opsize,lrc_packet op)
{ int goon;
  lrc_cell c,c1;
  int ipnum=ipsize/sizeof(lrc_ptr);
  int num_func_remv;
  lrc_cell new_cell;           /* Variable to store the function call */
  list_gen g;
  int mem,ag,t;
  lrc_bool found_in_new_cache = lrc_false;
  lrc_bool copied_old_gen = lrc_false;
  lrc_bool found_in_oldest = lrc_false;
  lrc_bool found_in_all    = lrc_false;

  lrc_memo_index i =  HASH(f);
  { int n;
    for(n=0; n<ipnum; n++) i=REHASH_IX_PTR(i,ip[n]);
  }


  #ifdef OLDEST_NEW_YOUNGEST
  { lrc_bool found = lrc_false;
    {                                /* Lookup oldest generation */
      c = LIST(i);
    #ifdef COMP_STATISTICS
      while ( c!= EMPTY && !found_in_oldest )
        { if ( EQ(f,ip,ipsize,c) )
            { found_in_oldest = lrc_true;
              stats_hit_in_old++;
            }
          else
            c = NEXT(c);
          stats_num_hash_op++;
        }
    #else
      while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
        c = NEXT(c);
      found_in_oldest = (c != EMPTY);
    #endif
    }

    if ( !found_in_oldest )
      { { /* fprintf(stdout," New "); */
          c = (lrc_cell) (*(GEN(l_generations) + i));
        #ifdef COMP_STATISTICS
          { lrc_bool enc = lrc_false;
            while ( c != EMPTY && !enc )
              { if ( EQ(f,ip,ipsize,c) )
                  { enc = lrc_true;
                    stats_hit_in_new++;
                  }
                else
                  c = NEXT(c);
                stats_num_hash_op++;
              }
          }
          if ( c!= EMPTY )
            { found = lrc_true;
              found_in_new_cache = lrc_true;
            }
        #else
          while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
            c = NEXT(c);
          found_in_new_cache = (c != EMPTY);
        #endif
        }

        if (!found_in_new_cache)
          { { if (l_younger_gen != EMPTY)        /* Lookup the youngest generation */
                { /* fprintf(stdout," Youngest "); */
                  c = (lrc_cell) (*(GEN(l_younger_gen) + i));
            #ifdef COMP_STATISTICS
                  while ( c!= EMPTY && !found_in_all )
                    { if ( EQ(f,ip,ipsize,c) )
                        { found_in_all = lrc_true;
                          stats_hit_in_all++;
                        }
                      else
                        c = NEXT(c);
                      stats_num_hash_op++;
                    }
            #else
                  while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
                    c = NEXT(c);
            #endif
                  found_in_all = (c != EMPTY);
                }
            }

          }
      }

    if ( stats_copying_to_old_gen && !found_in_oldest && !found_in_new_cache )
      { lrc_cell c2 = c;
        { found = lrc_true;

          for ( g = l_younger_gen ; g != EMPTY && found ; g = NEXT_GEN(g) )
            { c = (lrc_cell) *(GEN(g) + i);
        #ifdef COMP_STATISTICS
              { lrc_bool enc = lrc_false;
                while ( c != EMPTY && !enc )
                  { if ( EQ(f,ip,ipsize,c) )
                      enc = lrc_true;
                    else
                      c = NEXT(c);
                    stats_num_hash_op++;
                  }
              }
        #else
              while ( c != EMPTY && !EQ(f,ip,ipsize,c) )
                  c = NEXT(c);
        #endif
              found = (c != EMPTY);
            }

          if (found)
            { lrc_cell c2;
              /* fprintf(stdout," Copy to oldest generation!! "); */
              copied_old_gen = lrc_true;
              new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
              c2 = new_cell;
              *c2++ = LIST(i);
              *c2++ = (lrc_cell) 0;
              *c2++ = (lrc_cell) stats_num_allocs;
              memcpy(c2,c+3,ipsize + opsize + 2 * sizeof(lrc_cell));
              LIST(i) = new_cell;
              if ( NBE(i) >= MAX_NUMB_ENTRIES_PER_BUCKET )
                { replac_func_bucket_LRU(i);
        #ifdef COMP_STATISTICS
                  stats_num_replac++;
        #endif
                }
              else
                NBE(i)++;
              #ifdef COMP_STATISTICS
              stats_num_copy_op_old_gen++;
              stats_num_bytes_copied_old_gen += MEM(new_cell);
              #endif

            }
        }

        c = c2;
      }
  }

  #endif
  #ifdef YOUNGEST_OLDEST_NEW
  { lrc_bool found = lrc_false;

    { if (l_younger_gen != EMPTY)        /* Lookup the youngest generation */
        { /* fprintf(stdout," Youngest "); */
          c = (lrc_cell) (*(GEN(l_younger_gen) + i));
    #ifdef COMP_STATISTICS
          while ( c!= EMPTY && !found_in_all )
            { if ( EQ(f,ip,ipsize,c) )
                { found_in_all = lrc_true;
                  stats_hit_in_all++;
                }
              else
                c = NEXT(c);
              stats_num_hash_op++;
            }
    #else
          while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
            c = NEXT(c);
    #endif
          found_in_all = (c != EMPTY);
        }
    }

    if (!found_in_all)
      { {                                /* Lookup oldest generation */
          c = LIST(i);
        #ifdef COMP_STATISTICS
          while ( c!= EMPTY && !found_in_oldest )
            { if ( EQ(f,ip,ipsize,c) )
                { found_in_oldest = lrc_true;
                  stats_hit_in_old++;
                }
              else
                c = NEXT(c);
              stats_num_hash_op++;
            }
        #else
          while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
            c = NEXT(c);
          found_in_oldest = (c != EMPTY);
        #endif
        }

        if (!found_in_oldest)
          { /* fprintf(stdout," New "); */
            c = (lrc_cell) (*(GEN(l_generations) + i));
          #ifdef COMP_STATISTICS
            { lrc_bool enc = lrc_false;
              while ( c != EMPTY && !enc )
                { if ( EQ(f,ip,ipsize,c) )
                    { enc = lrc_true;
                      stats_hit_in_new++;
                    }
                  else
                    c = NEXT(c);
                  stats_num_hash_op++;
                }
            }
            if ( c!= EMPTY )
              { found = lrc_true;
                found_in_new_cache = lrc_true;
              }
          #else
            while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
              c = NEXT(c);
            found_in_new_cache = (c != EMPTY);
          #endif
          }

      }

    if ( stats_copying_to_old_gen && !found_in_oldest )
     { lrc_cell c2 = c;
        { found = lrc_true;

          for ( g = l_younger_gen ; g != EMPTY && found ; g = NEXT_GEN(g) )
            { c = (lrc_cell) *(GEN(g) + i);
        #ifdef COMP_STATISTICS
              { lrc_bool enc = lrc_false;
                while ( c != EMPTY && !enc )
                  { if ( EQ(f,ip,ipsize,c) )
                      enc = lrc_true;
                    else
                      c = NEXT(c);
                    stats_num_hash_op++;
                  }
              }
        #else
              while ( c != EMPTY && !EQ(f,ip,ipsize,c) )
                  c = NEXT(c);
        #endif
              found = (c != EMPTY);
            }

          if (found)
            { lrc_cell c2;
              /* fprintf(stdout," Copy to oldest generation!! "); */
              copied_old_gen = lrc_true;
              new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
              c2 = new_cell;
              *c2++ = LIST(i);
              *c2++ = (lrc_cell) 0;
              *c2++ = (lrc_cell) stats_num_allocs;
              memcpy(c2,c+3,ipsize + opsize + 2 * sizeof(lrc_cell));
              LIST(i) = new_cell;
              if ( NBE(i) >= MAX_NUMB_ENTRIES_PER_BUCKET )
                { replac_func_bucket_LRU(i);
        #ifdef COMP_STATISTICS
                  stats_num_replac++;
        #endif
                }
              else
                NBE(i)++;
              #ifdef COMP_STATISTICS
              stats_num_copy_op_old_gen++;
              stats_num_bytes_copied_old_gen += MEM(new_cell);
              #endif

            }
        }

        c = c2;
      }
  }

  #endif
  #ifdef NEW_OLDEST_YOUNGEST
  { lrc_bool found = lrc_false;

    { /* fprintf(stdout," New "); */
      c = (lrc_cell) (*(GEN(l_generations) + i));
    #ifdef COMP_STATISTICS
      { lrc_bool enc = lrc_false;
        while ( c != EMPTY && !enc )
          { if ( EQ(f,ip,ipsize,c) )
              { enc = lrc_true;
                stats_hit_in_new++;
              }
            else
              c = NEXT(c);
            stats_num_hash_op++;
          }
      }
      if ( c!= EMPTY )
        { found = lrc_true;
          found_in_new_cache = lrc_true;
        }
    #else
      while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
        c = NEXT(c);
      found_in_new_cache = (c != EMPTY);
    #endif
    }

    if ( !found_in_new_cache )
      { {                                /* Lookup oldest generation */
          c = LIST(i);
        #ifdef COMP_STATISTICS
          while ( c!= EMPTY && !found_in_oldest )
            { if ( EQ(f,ip,ipsize,c) )
                { found_in_oldest = lrc_true;
                  stats_hit_in_old++;
                }
              else
                c = NEXT(c);
              stats_num_hash_op++;
            }
        #else
          while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
            c = NEXT(c);
          found_in_oldest = (c != EMPTY);
        #endif
        }

        if (!found_in_oldest)
          { found = lrc_false;
            { { if (l_younger_gen != EMPTY)        /* Lookup the youngest generation */
                  { /* fprintf(stdout," Youngest "); */
                    c = (lrc_cell) (*(GEN(l_younger_gen) + i));
              #ifdef COMP_STATISTICS
                    while ( c!= EMPTY && !found_in_all )
                      { if ( EQ(f,ip,ipsize,c) )
                          { found_in_all = lrc_true;
                            stats_hit_in_all++;
                          }
                        else
                          c = NEXT(c);
                        stats_num_hash_op++;
                      }
              #else
                    while ( c!= EMPTY && !EQ(f,ip,ipsize,c) )
                      c = NEXT(c);
              #endif
                    found_in_all = (c != EMPTY);
                  }
              }

            }
            if ( stats_copying_to_old_gen )
              { lrc_cell c3 = c;
                { found = lrc_true;

                  for ( g = l_younger_gen ; g != EMPTY && found ; g = NEXT_GEN(g) )
                    { c = (lrc_cell) *(GEN(g) + i);
                #ifdef COMP_STATISTICS
                      { lrc_bool enc = lrc_false;
                        while ( c != EMPTY && !enc )
                          { if ( EQ(f,ip,ipsize,c) )
                              enc = lrc_true;
                            else
                              c = NEXT(c);
                            stats_num_hash_op++;
                          }
                      }
                #else
                      while ( c != EMPTY && !EQ(f,ip,ipsize,c) )
                          c = NEXT(c);
                #endif
                      found = (c != EMPTY);
                    }

                  if (found)
                    { lrc_cell c2;
                      /* fprintf(stdout," Copy to oldest generation!! "); */
                      copied_old_gen = lrc_true;
                      new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
                      c2 = new_cell;
                      *c2++ = LIST(i);
                      *c2++ = (lrc_cell) 0;
                      *c2++ = (lrc_cell) stats_num_allocs;
                      memcpy(c2,c+3,ipsize + opsize + 2 * sizeof(lrc_cell));
                      LIST(i) = new_cell;
                      if ( NBE(i) >= MAX_NUMB_ENTRIES_PER_BUCKET )
                        { replac_func_bucket_LRU(i);
                #ifdef COMP_STATISTICS
                          stats_num_replac++;
                #endif
                        }
                      else
                        NBE(i)++;
                      #ifdef COMP_STATISTICS
                      stats_num_copy_op_old_gen++;
                      stats_num_bytes_copied_old_gen += MEM(new_cell);
                      #endif

                    }
                }

                c = c3;
              }
          }
      }
  }

  #endif


  if ( c == EMPTY )
    { /************ not in cache, compute & store it */
      stats_memory_used += CELL_SIZE(ipsize,opsize);
      f(ip,op);                           /* Call the function */
      new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
      c = new_cell;
      /* fprintf(stdout," NEWCELL: %x ",new_cell); */
      *c++ = *(GEN(l_generations) + i);            /* NEXT */
      *c++ = (lrc_cell) 0;                         /* HITS */
      *c++ = (lrc_cell) stats_num_allocs;          /* BIRTH */
      *c++ = (lrc_cell) CELL_SIZE(ipsize,opsize);  /* MEM */
      *(GEN(l_generations) + i) = new_cell;
      *c++ = (lrc_cell)f;                          /* FUNC */
      memcpy(c,ip,ipsize);
      c += ipnum;
      memcpy(c,op,opsize);
      stats_num_allocs++;
      stats_size_allocs += CELL_SIZE(ipsize,opsize);
      stats_num_entries++;
      return 0;


    }
  else
    { /****************************** in cache, retrieve it */
      stats_num_hits++;
      HITS(c)++;
      /* change the birth-time to reflect hit */
      BIRTH(c)=stats_num_allocs;
      memcpy(op,OP(c,ipnum),opsize);
      if ( !found_in_oldest && !copied_old_gen  &&  !found_in_new_cache )
      {
        new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
        c1 = new_cell;
        *c1++ = *(GEN(l_generations) + i);            /* NEXT */
        *c1++ = (lrc_cell) HITS(c);                   /* HITS */
        *c1++ = (lrc_cell) stats_num_allocs;          /* BIRTH */
        memcpy(c1,c+3,ipsize + opsize + 2 * sizeof(lrc_cell));
        *(GEN(l_generations) + i) = new_cell;
        #ifdef COMP_STATISTICS
        stats_num_copy_op++;
        stats_num_bytes_copied += MEM(new_cell);
        #endif

      }


      return HITS(c);

    }
}

static int replac_func_bucket_LRU(int ind_bucket)
{ lrc_cell c,a,prev;
  int mem,bt;

  bt = BIRTH(LIST(ind_bucket));
  a = prev = EMPTY;
  for ( c = LIST(ind_bucket) ; c != EMPTY ; c = NEXT(c) )
    {  if ( BIRTH(c) < bt )
         { bt = BIRTH(c);
           prev = a;
         }
       a = c;
    }
  if ( prev == EMPTY ) /* replace the bucket's head */
    { mem = MEM(LIST(ind_bucket));
      LIST(ind_bucket) = NEXT(LIST(ind_bucket));
    }
  else
    { mem = MEM(NEXT(prev));
      NEXT(prev) = NEXT(NEXT(prev));
    }
  return mem;
}


int lrc_memo (lrc_visfun f,int ipsize,lrc_packet ip,int opsize,lrc_packet op)
{
  switch (memo_strateg)
    { case 0:                          /* No Memoing */
              (((lrc_visfun)f)((lrc_packet)ip,(lrc_packet)op));
              stats_num_allocs++;
              return 0;
              break;
      case 1: if (memo_tree_size)      /* Full Memoing */
                { if (lrc_term_nnodes(*ip) > stats_min_tree_size_to_cache)
                    return lrc_memo_full_memoing(f,ipsize,ip,opsize,op);
                  else
                    (((lrc_visfun)f)((lrc_packet)ip,(lrc_packet)op));
                }
              else
                return lrc_memo_full_memoing(f,ipsize,ip,opsize,op);
              break;
      case 2: if (memo_tree_size)      /* Delta Cache */
                { if (lrc_term_nnodes(*ip) > stats_min_tree_size_to_cache)
                    return lrc_memo_delta_cache(f,ipsize,ip,opsize,op);
                  else
                    (((lrc_visfun)f)((lrc_packet)ip,(lrc_packet)op));
                }
              else
                return lrc_memo_delta_cache (f,ipsize,ip,opsize,op);
              break;
      case 3: if (memo_tree_size)      /* Generations */
                { if (lrc_term_nnodes(*ip) > stats_min_tree_size_to_cache)
                    return lrc_memo_generations(f,ipsize,ip,opsize,op);
                  else
                    (((lrc_visfun)f)((lrc_packet)ip,(lrc_packet)op));
                }
              else
                return lrc_memo_generations (f,ipsize,ip,opsize,op);
    }
}

static int lrc_memo_full_memoing (lrc_visfun f,int ipsize,lrc_packet ip,
                                               int opsize,lrc_packet op)
{ int goon;
  lrc_cell c;
  int ipnum=ipsize/sizeof(lrc_ptr);

  lrc_memo_index i =  HASH(f);
  { int n;
    for(n=0; n<ipnum; n++) i=REHASH_IX_PTR(i,ip[n]);
  }


  c = *(GEN(l_generations) + i);
  #ifdef COMP_STATISTICS
  { lrc_bool found = lrc_false;
    while (c!=EMPTY && !found)
    { if ( EQ(f,ip,ipsize,c) )
        found = lrc_true;
      else
        c=NEXT(c);
      stats_num_hash_op++;
    }
  }
  #else
  while ( c!=EMPTY && !EQ(f,ip,ipsize,c) )
    c = NEXT(c);
  #endif

  if ( c == EMPTY )
    { /************ not in cache, compute & store it */
      lrc_cell new_cell;
      f(ip,op);                           /* Call the function */
      new_cell = (lrc_cell)ALLOC_CELL(CELL_SIZE(ipsize,opsize));
      c = new_cell;
      *c++ = *(GEN(l_generations) + i);            /* NEXT */
      *c++ = (lrc_cell) 0;                         /* HITS */
      *c++ = (lrc_cell) stats_num_allocs;          /* BIRTH */
      *c++ = (lrc_cell) CELL_SIZE(ipsize,opsize);  /* MEM */
      *(GEN(l_generations) + i) = new_cell;
      *c++ = (lrc_cell)f;                          /* FUNC */
      memcpy(c,ip,ipsize);
      c += ipnum;
      memcpy(c,op,opsize);
      stats_num_allocs++;
      stats_size_allocs += CELL_SIZE(ipsize,opsize);
      stats_memory_used += CELL_SIZE(ipsize,opsize);
      stats_num_entries++;
      #ifdef FUNC_HIST
      lrc_memo_f_set_calls(lrc_vpr_index(f));
      #endif

      return 0;

    }
  else
    { /****************************** in cache, retrieve it */
      stats_num_hits++;
      HITS(c)++;
      /* change the birth-time to reflect hit */
      BIRTH(c)=stats_num_allocs;
      memcpy(op,OP(c,ipnum),opsize);
      #ifdef FUNC_HIST
      lrc_memo_f_set_calls(lrc_vpr_index(f));
      lrc_memo_f_set_hits(lrc_vpr_index(f));
      #endif

      #ifdef TREE_SIZE_HIST
      lrc_memo_set_hit_ts (lrc_term_nnodes(*ip));
      #endif



      return HITS(c);

    }
}


int  lrc_memo_time() {
  return stats_num_allocs;
}

int  lrc_memo_hits(lrc_cell c) {
  return HITS(c);
}

int  lrc_memo_birth(lrc_cell c) {
  return BIRTH(c);
}

lrc_visfun lrc_memo_visfun(lrc_cell c) {
  return FUNC(c);
}


static int birth_base=0;
void lrc_memo_dump_set_birth_time(int birth_time) {
  birth_base=birth_time;
  return;
}


void lrc_memo_dump(lrc_memo_filter filter,lrc_memo_printer printer)
{
  lrc_memo_index i;
  list_gen g;
  int numb_entries,memory_used,numb_hits;

  for (g = l_generations ; g != EMPTY ; g = NEXT_GEN(g) )
    { numb_entries = 0;
      memory_used = 0;
      numb_hits = 0;
      for (i=0 ; i<lrc_memo_tablesize ; i++)
        { lrc_cell c; int n=0; int m=0;
          for (c=*(GEN(g) + i) ; c!=EMPTY ; c=NEXT(c),n++)
            { numb_entries++;
              memory_used += MEM(c);
              numb_hits   += HITS(c);
              if (filter(c) && BIRTH(c)>=birth_base) {
                printf(" %x MEMO %d %d ",c,HITS(c),BIRTH(c)); fflush(stdout);
                printer(c);  printf("\n");
              }
              else m++;
             /*   if (n>m)
            printf("BUCKET %d  list= %d hide= %d TOTAL= %d\n\n",i,n-m,m,n); */
            }
        }
      fprintf(stdout," Entries: %d   Memory: %d  Hits: %d \n",
              numb_entries,memory_used,numb_hits);
    }

#ifdef COMP_STATISTICS
  fprintf (stdout," Lookup Oper.: %d  ",stats_num_hash_op);
  fprintf (stdout," Copy Oper.: %d   ",stats_num_copy_op);
  fprintf (stdout," Bytes Copied: %d \n",stats_num_bytes_copied);
  fprintf (stdout," Copy Oper. to old gen: %d  ",
           stats_num_copy_op_old_gen);
  fprintf (stdout," Bytes Copied to old gen: %d \n",
           stats_num_bytes_copied_old_gen);

  fprintf (stdout," HITS Old: %d Youngest: %d New: %d \n",stats_hit_in_old,
           stats_hit_in_all,stats_hit_in_new);
  fprintf (stdout," Replacements in old: %d \n",stats_num_replac);
#endif

  fprintf (stdout," Total of Memory Allocated: %d  Array D'hashing: %d  Entries: %d \n",stats_size_allocs,lrc_memo_tablesize * sizeof(lrc_ptr),numb_entries);
  return;
}

void lrc_memo_statistics(
   int * num_allocs,
   int * num_hits
) {
  *num_allocs=stats_num_allocs;
  *num_hits=stats_num_hits;
}

static float root(float x) {
  float a=0,b=x,m;
  while ((a-b)*(a-b)>1E-08) { m=(a+b)/2; if (m*m<x) a=m; else b=m; }
  return (a+b)/2;
}

static void stats(int lev,float *aver,float *sdev,int *longest,int *shortest) {
  lrc_memo_index i; float av,var; int min; int max;

  av=(float)stats_num_entries/(float)lrc_memo_tablesize;
  var=0; min=stats_num_entries; max=0;

  for(i=0; i<lrc_memo_tablesize; i++) {
    lrc_cell t; int n;
    for(t= *(GEN(l_generations)+i),n=0; t!=EMPTY; t=NEXT(t),n++) ;
    if (lev>3) printf("%d : %d\n",i,n);
    var+=(av-n)*(av-n);
    if (n>max) max=n;
    if (n<min) min=n;
  }

  var=var/(float)(lrc_memo_tablesize-1);
  *aver=av; *sdev=root(var); *longest=max; *shortest=min;

  return;
}

void lrc_memo_stats(int level) {
  float av; float sd; int longest; int shortest;

  printf("MEMO   n=%d  misses=%d  hits=%d  total=%d\n\n",lrc_memo_tablesize,
    stats_num_allocs,stats_num_hits,stats_num_allocs+stats_num_hits);
  if (level==0) return;

  printf("SIZE   total=%d  admin=%d  user=%d\n",
    stats_size_allocs,
    stats_num_allocs*ADMIN_SIZE*sizeof(lrc_ptr),
    stats_size_allocs-stats_num_allocs*ADMIN_SIZE*sizeof(lrc_ptr));
  if (level==1) return;

  stats(level,&av,&sd,&longest,&shortest);
/*
  printf("LENGTH Min=%d  Av=%.1f  Max=%d  Sd=%.2f\n",shortest,av,longest,sd);
  if (level==2) return;

  printf("84%% of the lists is shorter than %.1f (=Av+%.1f%%)\n"
        ,av+sd,sd/av*100);
  printf("98%% of the lists is shorter than %.1f (=Av+%.1f%%)\n"
        ,av+2*sd,2*sd/av*100); */
  /* in a normal distribution, 68% lies in av+/-sd  */
  /*                       and 95% lies in av+/-2sd */
  return;
}




