#include <stdio.h>
#include "lrcmemory.h"
#include "prims.h"
#include "initial.h"
#include "grammar_tables.h"
#include "reference.h"


#define bump(p) p=p->next
#define empty_path 0
#define is_empty_path(p) (p==0)

#define new_path_elem (PathElement *)lrc_malloc(sizeof(PathElement))

Path create_root_path(lrc_term t) {
  Path res;
  res=empty_path;
  return res;
}

int path_is_root_path(Path p) {
  return is_empty_path(p);
}

int path_is_singleton(Path p) {
  if (is_empty_path(p)) return 0;
  return is_empty_path(p->next);
}

int path_num_edges(Path p) {
  int res;
  res=0;
  while (p!=0) { res++; bump(p); }
  return res;
}

Path path_extend(Path p, int child_number) {
  Path newp;

  newp=new_path_elem;
  newp->childnumber=child_number;
  newp->next=p;
  return newp;
}

int path_equal(Path p1, Path p2) {
  while (p1!=0 && p2!=0) {
    if (! (p1->childnumber==p2->childnumber))
              return 0;
    bump(p1);
    bump(p2);
  }
  return (p1==0 && p2==0);
}

Path path_copy(Path p) {
  Path newp,newproot;

  if (is_empty_path(p)) return empty_path;
  newproot=new_path_elem;
  newproot->childnumber=p->childnumber;
  for (newp=newproot; p->next!=0;) {
    bump(p);
    newp->next=new_path_elem;
    bump(newp);
    newp->childnumber=p->childnumber;
  }
  newp->next=empty_path;
  return newproot;
}

Path path_concat(Path p1, Path p2) {
  Path newp1,newp2,newp1root;

  newp2=path_copy(p2);                  /* copy p2 => newp2  */
  if (p1==0)
          return newp2;
  newp1root=new_path_elem;
  newp1root->childnumber=p1->childnumber;
  for (newp1=newp1root; p1->next!=0;) { /* copy p1 => newp1 */
    bump(p1);
    newp1->next=new_path_elem;
    bump(newp1);
    newp1->childnumber=p1->childnumber;
  }
  newp1->next=newp2;                    /* connect newp2 to newp1 */
  return newp1root;
}

Path  path_take(Path p, int num_edges) {
  Path res, pp;
  if (p==0) return 0;
  res=path_copy(p);
  pp=res;
  /* perform (num_edges-1) bumps */
  while (num_edges-1>0) { bump(pp); num_edges--; }
  pp->next=0;
  return res;
}

Path  path_remove_deepest(Path p, int num_edges) {
  Path res;
  int n; /* number of edges in resulting path */

  if (p==0) return 0;
  n=path_num_edges(p)-num_edges;
  if (n<=0) return 0;

  res=path_take(p,n);
  return res;
}

int path_is_ancestor(Path p1, Path p2) {
  while (p1!=0 && p2!=0) {
    if (p1->childnumber!=p2->childnumber) return 0;
    bump(p1);
    bump(p2);
  }
  return (p1==0);
}

Path path_clip(Path anc, Path p) {
  Path q;
  q=p;
  while (anc!=0) {
      bump(anc); bump(q);
  }
  return path_copy(q);
}

Path path_common_ancestor(Path p1, Path p2) {
  Path newp;
        
  if (p1==0 || p2==0) return 0;
  
  if (p1->childnumber==p2->childnumber) {
    newp=path_common_ancestor(p1->next,p2->next);
    newp=path_extend(newp,p1->childnumber);
  }
  else newp=empty_path;
  return newp;
}

void path_print(Path p) {
  if (p==0) { printf("/"); return; }
  while (p!=0) {
    printf("/%d",p->childnumber);
    bump(p);
  }
}

lrc_term path_map_to_term(lrc_term root, Path p) {
  lrc_term t;

  if (p==0) return root; /* path is root */

  t=root;
  while(p!=0 && p->childnumber!=0) {
    if (lrc_constructor_num_sons(LRC_TAG(t)) < p->childnumber) return 0;
    t=LRC_CHILD(t,p->childnumber);
    bump(p);
  }
  return t;
}

lrc_term path_deref_length(lrc_term root ,Path p, int path_length) {
  lrc_term t;

  t=root;
  while (path_length>0) {
    if (lrc_constructor_num_sons(LRC_TAG(t)) < p->childnumber) return 0;
    t=LRC_CHILD(t,p->childnumber);
    bump(p);
    path_length--;
  }
  return t;
}

Path term_map_to_path(lrc_term root,lrc_term t) {
    Path p;
    int goon;

    if (lrc_is_initial(t)==lrc_false) {
       fprintf(stderr,"term_map_to_path: t is not an initial term \n");
       lrc_term_print_long(t); printf("\n");
       return 0;
    }
    p=0;
    goon=1;
    while(goon && (t!=root) && (t!=0)) {
       lrc_term parent;
       int child;
       parent=lrc_initial_parent(t);
       goon=parent!=0;
       if (parent!=0) {
         child=lrc_initial_index_parent(t);
         p=path_extend(p,child);
         t=lrc_initial_parent(t);
       }
    }
    if (t==0) {
      fprintf(stderr,"severe error:term_map_to_path could not map term\n");
    }
    return p;
}

Reference new_reference() {
  Reference res;
  res=(Reference)lrc_malloc(sizeof(*res));
  return res;
}

Reference create_root_reference(lrc_term root) {
  Reference res;

  res=new_reference();
  res->tree=root;
  res->kind=REF_SUBTREE;
  res->sub0=create_root_path(root);
  res->sub1=0;
  return res;
}

Reference subterm_to_reference(lrc_term t) {
  Reference res;
  lrc_term t_root;
  t_root=lrc_initial_root(t);
  res=new_reference();
  res->tree=t_root;
  res->kind=REF_SUBTREE;
  res->sub0=term_map_to_path(t_root,t);
  res->sub1=0;
  return res;
}

Reference reference_deref(Reference r) {
  Reference res;

  res=new_reference();
  res->kind=REF_SUBTREE;
  res->tree=path_map_to_term(r->tree,r->sub0);
  res->sub0=create_root_path(res->tree);
  return res;
}

lrc_term reference_to_term(Reference r) {
  return path_map_to_term(r->tree,r->sub0);
}

Reference term_to_reference(lrc_term root, lrc_term t) {
  Reference res;
  
  /* TODO: check term_to_reference */
  res=new_reference();
  res->kind=REF_SUBTREE;
  res->tree=root;
  res->sub0=term_map_to_path(root,t);
  return res;
}


int references_overlap(int num_refs, Reference * refs) {
  int ix;
  Path path, new_path;

   path=refs[0]->sub0;
   for(ix=1;ix<num_refs;ix++) {
      new_path=path_common_ancestor(path,refs[ix]->sub0);
      if (path_equal(new_path,path) || path_equal(new_path,refs[ix]->sub0) ) return 1;
      path=new_path;
   }
   return 0;
}


