/*
** strs.c
** march 1994 by Maarten Pennings
** Library functions by Matthijs Kuiper
*/

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include "strs.h"

#define STR_INDEX_OFFSET 300
static lrc_term_ix index_str(lrc_string s) {
  lrc_term_ix ix;
  /* strange but lrc_true: chars can be negative */
  for(ix=T__Str+STR_INDEX_OFFSET; *s!=0; s++) ix = (2*ix + abs(*s)) % lrc_term_tabelsize;
  return ix;
}

static lrc_bool eq_str(lrc_term t,lrc_ptr b) {
  return LRC_TAG(t)==FIELD(b,0,int) && 
    strcmp(LRC_VALUE_lrc_string(t),FIELD(b,sizeof(lrc_ptr),lrc_string))==0;
}

static lrc_term lrc_empty_string;

static lrc_term C_Str_local(lrc_string s) {
  struct {lrc_ptr tag;lrc_string s;} buf;
  buf.tag=(lrc_ptr)T__Str; buf.s=s;
  return lrc_shared_prim(index_str(s),eq_str,sizeof(buf),&buf);
}

lrc_term C_Str(lrc_string s) {
  if (*s=='\0') return lrc_empty_string;
  else {
    struct {lrc_ptr tag;lrc_string s;} buf;
    buf.tag=(lrc_ptr)T__Str; buf.s=s;
    return lrc_shared_prim(index_str(s),eq_str,sizeof(buf),&buf);
  }
}

/*********************\
** Library functions **
\*********************/


int lrc_string_compare(lrc_string a, lrc_string b) {
  int cmp=strcmp(a,b);
  if (cmp<0) return lrc_less;
  if (cmp==0) return lrc_equal;
  return lrc_greater;
}

int lrc_string_compare_term(lrc_term a, lrc_term b) {
  return lrc_string_compare(LRC_VALUE_lrc_string(a),LRC_VALUE_lrc_string(b));
}

lrc_bool lrc_string_equal(lrc_string a, lrc_string b) {
  return !strcmp(a,b);
}

lrc_bool lrc_string_equal_term(lrc_term a, lrc_term b) {
  return lrc_string_equal(LRC_VALUE_lrc_string(a),LRC_VALUE_lrc_string(b));
}

void lrc_string_fprint(FILE *f, lrc_string s) {
  while (*s!=0) lrc_char_fprint(f,*s++);
}

void lrc_string_print(lrc_string s) {
  lrc_string_fprint(stdout,s);
}

lrc_string lrc_string_const(lrc_string s) {
  return s;
}

lrc_string lrc_string_concat(lrc_string a, lrc_string b) {
  lrc_string res=(lrc_string)ALLOC_STR(strlen(a)+strlen(b)+1);
  *res='\0';
  /* printf("concat strings, <%s> <%s>\n", a,b); */
  strcat(res,a);strcat(res,b);
  /* printf("res=%s\n",res); */
  return res;
}

lrc_string lrc_vstring_concat(lrc_string s,...) {
  va_list ap;
  int num_bytes=strlen(s);
  lrc_string st;
  lrc_string res;

  va_start(ap,s);
  st=va_arg(ap,lrc_string);
  while (st!=0) {
    num_bytes+=strlen(st);
    st=va_arg(ap,lrc_string);
  }
  va_end(ap);
  
  res=(lrc_string)ALLOC_STR(num_bytes+1);
  *res='\0';
  va_start(ap,s);
  strcat(res,s);
  st=va_arg(ap,lrc_string);
  while (st!=0) {
    strcat(res,st);
    st=va_arg(ap,lrc_string);
  }
  va_end(ap);
  return res;
}

lrc_string lrc_string_sub(lrc_string s, int lwb, int upb) {
  int start_ix, end_ix; /*indexes in s*/
  int lg ; /*length of s*/
  lg = strlen(s);
  if (lwb<1)  start_ix = 1; else start_ix = lwb;
  if (upb>lg) end_ix = lg; else end_ix = upb;
  if (end_ix<start_ix) return "";
  else {
    int num_chars=end_ix+1-start_ix;
    lrc_string res=(lrc_string)ALLOC_STR(num_chars+1);
    lrc_string start=s+(start_ix-1);
    *res='\0';
    strncpy(res,start,num_chars);
    res[num_chars]='\0';
    return res;
  }
}

lrc_string lrc_string_opensub(lrc_string s, int ix) {
  return lrc_string_sub(s,ix ,strlen(s));
}

char lrc_string_index(lrc_string s, int ix) {
   return s[ix-1];
}

char * lrc_string_copy(char *s) {
  char * res;
  res = lrc_malloc(strlen(s)+1);
  strcpy(res,s);
  return res;
}


static lrc_string string_default;

lrc_string lrc_string_default() {
  return string_default;
}

void lrc_Str_init() {
  lrc_empty_string=C_Str_local("");
  string_default=(lrc_string)ALLOC_STR(sizeof(*string_default));
  *string_default='\0';
}
