#include <stdio.h>
#include <string.h>
#include "shalloc.h"
#include "strs.h"
#include "maps.h"
#include "style.h"



StyleList add_style(Style head, StyleList l) {
  StyleList res;
  res=(StyleList)lrc_malloc(sizeof(*res));
  res->next=l;
  res->element=head;
  return res;
}

Style get_style(lrc_string style_name, StyleList lst) {
  while (lst!=0 && !lrc_string_equal(lst->element->name,style_name)) {
    lst=lst->next;
  }
  if (lst==0) return 0;
  else return lst->element;
}

Style style_create(lrc_string name) {
  Style res;
  res=(Style)lrc_malloc(sizeof(*res));
  res->name=name;
  res->properties
  = lrc_empty_map(
      lrc_string_equal_term,
      lrc_string_compare_term,
      lrc_string_equal_term,
      lrc_string_compare_term,C_Str("")
    ); 
  return res;
}

void style_define_property(Style st, lrc_string prop, lrc_string value) {
  /* printf("style_define_property: <%s> <%s>\n",prop,value); */
  st->properties=lrc_map_single_update(C_Str(prop), C_Str(value), st->properties);
}

lrc_string style_lookup(Style st, lrc_string prop_name) {
  lrc_term s;
  s=lrc_map_application(C_Str(prop_name),st->properties);
  return LRC_VALUE_lrc_string(s);
}

int style_contains(Style st, lrc_string prop_name) {
  return lrc_map_in(C_Str(prop_name),st->properties);
}

lrc_string style_lookup_str_default(Style st, lrc_string search_key, lrc_string deflt) {
  lrc_term key_term;

  if (st==0) return deflt;
  key_term=C_Str(search_key);
  if (lrc_map_in(key_term,st->properties)) return style_lookup(st,search_key);
  return deflt;
}

int style_lookup_int(Style st, lrc_string search_key) {
  lrc_string s;

  s=style_lookup(st,search_key);
  if (lrc_string_equal(s,"")) return 0;
  return STRtoINT(s);
}

int  style_lookup_int_default(Style st, lrc_string search_key, int deflt) {
  lrc_term key_term;

  key_term=C_Str(search_key);
  if (lrc_map_in(key_term,st->properties)) return STRtoINT(style_lookup(st,search_key));
  return deflt;
}

Style style_read_from_file(char * style_name, char * fname) {
  FILE *f;
  Style st;
  lrc_string line;

  f=fopen(fname,"r");
  if (!f) {
    /* printf("Cannot open %s\n",fname); */
    return 0;
  }
  st=style_create(style_name);
  line=(char *)lrc_malloc(401);
  while(fgets(line,400,f)) {
     /* printf("%s", line); fflush(stdout); */
     int len;
     len=strlen(line);
     if (len>2) {
        int space_index;
        int value_index;
        if (line[len-1]=='\n') line[len-1]='\0';
        space_index=strcspn(line," \t");
        value_index=space_index+strspn(line+space_index," \t");
      /*****
        printf("0123456789012345678901234567890\n");
        printf("%s\n",line);
        printf("  space_index=%d\n",space_index);
        printf("  value_index=%d\n",value_index);
       ******/
        if (space_index<len && value_index<len) {
           lrc_string prop, val;

           line[space_index]='\0';
           prop=lrc_string_concat(line,"");
           val=lrc_string_concat(line+value_index,"");
           /****
           printf("setting style property: <%s> -> <%s>\n",
                   prop,val);
           ****/
           style_define_property(st,prop,val);
        }
     }
  }
  return st;
}

