#ifndef TRACE_HANDLING
#define TRACE_HANDLING 0
#endif


#include <stdio.h>
#include <string.h>
#include "lrcassert.h"
#include "lrcmemory.h"
#include "prims.h"
#include "term.h"
#include "initial.h"
#include "reference.h"
#include "grammar_tables.h"
#include "surgery.h"

Replacement create_replacement(lrc_term t, int num_repl) {
  Replacement res;
  res=(Replacement)lrc_malloc(sizeof(*res));

  res->term = t;
  res->num_repl      = num_repl;
  res->operands      = (Reference *)lrc_malloc(num_repl * sizeof(*(res->operands)));
  res->replacements  = (lrc_term  *)lrc_malloc(num_repl * sizeof(lrc_term) );
  return res;
};

static lrc_term shared_replace(lrc_term t, Path path, lrc_term replacement) {
  if (TRACE_HANDLING) {
    printf("START replace: ");
    lrc_term_print_long(t); printf("/"); path_print(path);
    printf("<--");
    lrc_term_print_long(replacement);
    printf("\n");
  }

  if (path==0) return lrc_shadow(replacement);
  if (path_is_singleton(path)) {
     return shadow_replace_child(t,path->childnumber,replacement);
  }
  else {
    lrc_term rest;
    rest=shared_replace(LRC_CHILD(t,path->childnumber),path->next,replacement);
    return shadow_replace_child(t,path->childnumber,rest);
  }
}

lrc_term multi_replace(
  lrc_term ast, int num_repl, Reference * operands, lrc_term * replacements) {

  lrc_term result;
  int ix;

  result=lrc_shadow(ast);
  ix=0;
  while (ix<num_repl) {
     if (TRACE_HANDLING) {
       printf("MULTI REPLACE. Performing replacement number %d\n", ix);
       printf("PATH="); path_print(operands[ix]->sub0); printf("\n");
       printf("REPLACEMENT(%s)=", lrc_constructor_ident(LRC_TAG(replacements[ix])) );
       if (LRC_TAG(replacements[ix])==lrc_prod_name_to_tag("_Ptr") ) {
         printf(" A PTR TO ");
         lrc_term_print_long(LRC_VALUE_lrc_ptr(replacements[ix]));
       }
       else {
         lrc_term_print_long(replacements[ix]); 
       }
       printf("\n\n");
     }

     result=shared_replace(result,operands[ix]->sub0, replacements[ix]);
     ix++;
  }
  Assert(!lrc_is_initial(result),"multi_replace: result ast IS initial");
  return lrc_create_initial(result);
}

lrc_term in_situ_replace(lrc_term t, Path path, lrc_term replacement) {
  
  if (path==0) return replacement;
  if (path_is_singleton(path)) {
     return initial_replace_child(t,path->childnumber,replacement);
  }
  else {
    lrc_term rest;
    rest=in_situ_replace(LRC_CHILD(t,path->childnumber),path->next,replacement);
    return initial_replace_child(t,path->childnumber,rest);
  }
}

lrc_term in_situ_multi_replace
(
  lrc_term ast,
  Replacement replacement
)
{
  int num_repl;
  Reference * operands;
  lrc_term * replacements;
  lrc_term result;
  int ix;

  Assert(lrc_is_initial(ast),"IN in_situ_multi_replace: ast is NOT initial");
  operands=replacement->operands;
  replacements=replacement->replacements;
  num_repl=replacement->num_repl;

  if (references_overlap(num_repl, operands)) {
    int ix;
    printf("????\nIN performing replacements: operands overlap\n");
    printf("Replacement is not performed\n");
    printf("Here are the operands\n");
    for(ix=0;ix<num_repl;ix++) {
        printf("operand[%d]=", ix); path_print(operands[ix]->sub0);
        printf("==");
        lrc_term_print_long(reference_to_term(operands[ix]));
        printf("\n");
    }
    return ast;
  }


  result=ast;
  ix=0;
  while (ix<num_repl) {
     lrc_term initial_replacement;
     
     if (TRACE_HANDLING) {
       printf("COMPUTING initial_replacement IN in_situ_multi_replace\n");
     }
     initial_replacement=lrc_create_initial(replacements[ix]);
     if (TRACE_HANDLING) {
       printf("DONE COMPUTING initial_replacement IN in_situ_multi_replace\n");
     }
     result=in_situ_replace(result,operands[ix]->sub0, initial_replacement);
     ix++;
  }
  return result;
}


