#include <stdio.h>
#include <string.h>
#include "shalloc.h"
#include "initial.h"
#include "prims.h"
#include "strs.h"
#include "grammar_tables.h"
#include "textbuffer.h"
#include "position.h"
#include "unparse_tables.h"
#include "view_tables.h"
#include "reference.h"
#include "position.h"
#include "view_tables.h"
#include "fragments.h"
#include "views.h"


#ifndef TRACE_UNPARSER
#define TRACE_UNPARSER 0
#endif

#define TRACE_START if (TRACE_UNPARSER) {
#define TRACE_END }

#ifndef TRACE_POINTING
#define TRACE_POINTING 0
#endif

#define TRACE_POINTING_START if (TRACE_POINTING) {
#define TRACE_POINTING_END }


#define MIN_TEXT_BLOCK_SIZE 80

static TextBlockPair tb_split(TextBlock tb, int location) {
  TextBlockPair res;
  
  return res;
}

static TextBlock tb_create(void) {
  TextBlock res;
  res=(TextBlock)lrc_malloc(sizeof(*res));
  res->next=0;
  res->prev=0;
  res->text_buffer=textbuffer_init();
  return res;
}

static TextPieces tp_create(void) {
  TextPieces res;
  res=(TextPieces)lrc_malloc(sizeof(*res));
  res->first=0;
  res->last=0;
  return res;
}

static void tp_append_block(TextPieces pieces, TextBlock b) {
   if (pieces->first==0) {
       pieces->first=b;
       pieces->last=b;
       return;
   }
   pieces->last->next=b;
   b->prev=pieces->last;
   pieces->last=b;
}

ViewInstance create_view_instance(void) {

  ViewInstance res;

  res=(ViewInstance)lrc_malloc(sizeof(*res));
  res->term_ref=0;
  res->attributes=0;
  res->view_table.view_name=0;
  res->view_table.view_prod_table=0;
  res->view_table.view_unparse_commands=0;
  res->fragments=create_fragments();
  return res;
}

void vi_clear_unparsing(ViewInstance vi) {
  fragments_delete_elements(vi->fragments);
}

void vi_set_reference(ViewInstance vi, Reference r) {
  vi->term_ref=r;
}
Reference vi_get_reference(ViewInstance vi) {
  return vi->term_ref;
}

void vi_set_attributes(ViewInstance vi, lrc_term attrs) {
  vi->attributes=attrs;
}
void vi_set_view(ViewInstance vi, lrc_views_entry_tp table) {
  vi->view_table=table;
}

void vi_enter_term(ViewInstance vi, lrc_term t) {
   if (lrc_is_initial(t)) {
     vi_set_reference(vi,subterm_to_reference(t));
   }
   else vi_set_reference(vi,create_root_reference(t));
}

char * vi_get_view_name(ViewInstance vi) {
  return vi->view_table.view_name;
}

void vi_compute_fragments(ViewInstance vi) {
  int prod, num_prods;

  TRACE_START 
    printf("COMPUTING FRAGMENTS\n");
  TRACE_END
  if (vi->view_table.view_unparse_commands==0) {
     printf("vi_compute_fragments: view_table is empty");
     return;
  }

  num_prods=lrc_num_productions();

  TRACE_START
    printf("  NUM PRODUCTIONS is %d\n", num_prods);
  TRACE_END

  for(prod=0;prod<num_prods;prod++) {
    int command_index;
    int num_commands;
    int command_number;

    TRACE_START
      printf("COMPUTE FRAGMENTS. PROCESSING %s\n", lrc_constructor_ident(prod));
    TRACE_END
    if (lrc_prod_has_unparsing(vi->view_table,prod)) {
      command_index=lrc_prod_first_command_index(vi->view_table,prod);
      num_commands=lrc_prod_num_commands(vi->view_table,prod);
      for(command_number=1;command_number<=num_commands;command_number++) {
        unparse_commands_entry_tp command;

        command= lrc_prod_nth_command(vi->view_table,prod,command_number);
        {
          TRACE_START
            printf("processing string [%s]\n", command.command_data);
          TRACE_END
          fr_add_template_substring(command_index,0, strlen(command.command_data),vi->fragments);
        }


        command_index++;
      }
    }
  }
}

static unparse_commands_tp * vi_unparse_commands;
static int _index;
static void _print_element(UnparseElement e) {
  printf("    %d:",_index); print_unparse_element(e,vi_unparse_commands); printf("\n");
  _index++;
}

void vi_print(ViewInstance vi) {
  printf("VIEW INSTANCE\n");
  printf("  VIEW ");
  if (vi->view_table.view_name==0) {
    printf("<null>\n");
  }
  else {
    printf("%s\n",vi->view_table.view_name);
  }
  printf("  FRAGMENTS\n");
  vi_unparse_commands=vi->view_table.view_unparse_commands;
  _index=0;
  fragments_all_elements(vi->fragments,_print_element);
}


